from logging import getLogger

import psutil as ps

from ...commands.args import mm, WEB, READ, LOG, BASE, VERBOSITY, FORCE, DEV
from ...commands.read import STREAM, NAME, upload_files
from ...global_ import global_dev
from ...lib.log import Record
from ...lib.utils import parse_bool
from ...lib.workers import command_root

log = getLogger(__name__)


class Upload:

    def __init__(self, data):
        self.__data = data

    def __call__(self, request, s):
        files = [{NAME: file.filename, STREAM: file.stream} for file in request.files.getlist('files')]
        items = request.form.getlist('kit')
        force = parse_bool(request.form.get('force'))
        # we do this in two stages
        # first, immediate saving of files while web browser waiting for response
        upload_files(Record(log), self.__data, files=files, nfiles=len(files), items=items)
        # second, start rest of ingest process in background
        cmd = f'{command_root()} {mm(VERBOSITY)} 0 {mm(BASE)} {self.__data.base} {mm(LOG)} {WEB}-{READ}.log'
        if global_dev(): cmd += f' {mm(DEV)}'
        cmd += f' {READ}'
        if force: cmd += f' {mm(FORCE)}'
        log.info(f'Starting {cmd}')
        ps.Popen(args=cmd, shell=True)
        # wait so that the progress has time to kick in
        self.__data.sys.wait_for_progress(READ)




import datetime as dt

from IPython.core.display import display
from bokeh.io import output_file
from bokeh.layouts import row, gridplot
from bokeh.plotting import show

from ch2.names import N, like
from ch2.data import *
from ch2.lib import *
from ch2.pipeline.owners import *
from ch2.jupyter.decorator import template


@template
def compare_activities(local_time, compare_time, activity_group):

    f'''
    # Compare Activities: {local_time} v {compare_time} ({activity_group})
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')

    activity = std_activity_statistics(s, activity_journal=local_time, activity_group=activity_group)
    compare = std_activity_statistics(s, activity_journal=compare_time, activity_group=activity_group)
    health = std_health_statistics(s)
    hr_zones = hr_zones_from_database(s, local_time, activity_group)
    climbs = Statistics(s, activity_journal=local_time, activity_group=activity_group). \
        by_name(ActivityCalculator, N.ACTIVE_TIME, N.ACTIVE_DISTANCE). \
        by_name(ActivityCalculator, N.CLIMB_ANY, like=True).with_. \
        rename_with_units().df

    f'''
    ## Activity Plots
    
    The black line shows data from {local_time}, 
    the grey line from {compare_time}. 
    To the right of each plot of data against distance is a related plot of cumulative data
    (except the last, cadence, which isn't useful and so replaced by HR zones).
    Green and red areas indicate differences between the two dates. 
    Additional red lines on the altitude plot are auto-detected climbs.
    
    Plot tools support zoom, dragging, etc.
    '''

    output_file(filename='/dev/null')

    sp = comparison_line_plot(700, 200, N.DISTANCE_KM, N.MED_SPEED_KMH, activity, other=compare, ylo=0)
    add_climb_zones(sp, climbs, activity)
    sp_c = cumulative_plot(200, 200, N.MED_SPEED_KMH, activity, other=compare, ylo=0)
    xrange = sp.x_range if sp else None

    el = comparison_line_plot(700, 200, N.DISTANCE_KM, N.ELEVATION_M, activity, other=compare, x_range=xrange)
    add_climbs(el, climbs, activity)
    el_c = cumulative_plot(200, 200, N.CLIMB_MS, activity, other=compare)
    xrange = xrange or (el.x_range if el else None)

    hri = comparison_line_plot(700, 200, N.DISTANCE_KM, N.HR_IMPULSE_10, activity, other=compare, ylo=0, x_range=xrange)
    add_climb_zones(hri, climbs, activity)
    hri_c = cumulative_plot(200, 200, N.HR_IMPULSE_10, activity, other=compare, ylo=0)
    xrange = xrange or (hri.x_range if hri else None)

    hr = comparison_line_plot(700, 200, N.DISTANCE_KM, N.HEART_RATE_BPM, activity, other=compare, x_range=xrange)
    add_hr_zones(hr, activity, N.DISTANCE_KM, hr_zones)
    add_climb_zones(hr, climbs, activity)
    hr_c = cumulative_plot(200, 200, N.HEART_RATE_BPM, activity, other=compare)
    xrange = xrange or (hr.x_range if hr else None)

    pw = comparison_line_plot(700, 200, N.DISTANCE_KM, N.MED_POWER_ESTIMATE_W, activity, other=compare, ylo=0, x_range=xrange)
    add_climb_zones(pw, climbs, activity)
    pw_c = cumulative_plot(200, 200, N.MED_POWER_ESTIMATE_W, activity, other=compare, ylo=0)
    xrange = xrange or (pw.x_range if pw else None)

    cd = comparison_line_plot(700, 200, N.DISTANCE_KM, N.MED_CADENCE_RPM, activity, other=compare, ylo=0, x_range=xrange)
    add_climb_zones(cd, climbs, activity)
    hr_h = histogram_plot(200, 200, N.HR_ZONE, activity, xlo=1, xhi=5)

    show(gridplot([[el, el_c], [sp, sp_c], [hri, hri_c], [hr, hr_c], [pw, pw_c], [cd, hr_h]]))

    '''
    ## Activity Maps
    '''

    map = map_plot(400, 400, activity, other=compare)
    m_el = map_intensity_signed(200, 200, activity, N.GRADE_PC, ranges=map, power=0.5)
    m_sp = map_intensity(200, 200, activity, N.MED_SPEED_KMH, ranges=map, power=2)
    m_hr = map_intensity(200, 200, activity, N.HR_IMPULSE_10, ranges=map)
    m_pw = map_intensity(200, 200, activity, N.MED_POWER_ESTIMATE_W, ranges=map)
    show(row(map, gridplot([[m_el, m_sp], [m_hr, m_pw]], toolbar_location='right')))

    '''
    ## Activity Statistics
    '''

    '''
    Active time and distance exclude pauses.
    '''

    climbs[[N.ACTIVE_TIME_S, N.ACTIVE_DISTANCE_KM]].dropna(). \
        transform({N.ACTIVE_TIME_S: format_seconds, N.ACTIVE_DISTANCE_KM: format_km})

    '''
    Climbs are auto-detected and shown only for the main activity. They are included in the elevation plot above.
    '''

    if present(climbs, N.CLIMB_TIME):
        display(transform(climbs.filter(like='Climb').dropna(),
                          {N.CLIMB_TIME: format_seconds, N.CLIMB_ELEVATION: format_metres,
                           N.CLIMB_DISTANCE: format_km, N.CLIMB_GRADIENT: format_percent,
                           N.CLIMB_POWER: format_watts, N.CLIMB_CATEGORY: lambda x: x}))

    '''
    ## Health and Fitness
    '''

    fitness, fatigue = like(N.FITNESS_ANY, health.columns), like(N.FATIGUE_ANY, health.columns)
    colours = ['black'] * len(fitness) + ['red'] * len(fatigue)
    alphas = [1.0] * len(fitness) + [0.5] * len(fatigue)
    ff = multi_line_plot(900, 300, N.TIME, fitness + fatigue, health, colours, alphas=alphas)
    xrange = ff.x_range if ff else None
    add_multi_line_at_index(ff, N.TIME, fitness + fatigue, health, colours, alphas=alphas, index=-1)
    atd = std_distance_time_plot(900, 200, health, x_range=xrange)
    shr = multi_plot(900, 200, N.TIME, [N.DAILY_STEPS, N.REST_HR_BPM], health, ['grey', 'red'], alphas=[1, 0.5],
                     x_range=xrange, rescale=True, plotters=[bar_plotter(dt.timedelta(hours=20)), dot_plotter()])
    add_curve(shr, N.TIME, N.REST_HR_BPM, health, color='red', y_range_name=N.REST_HR_BPM)
    show(gridplot([[ff], [atd], [shr]]))




from logging import getLogger

from sqlalchemy.sql.functions import count

from .utils import Displayer
from ...diary.model import value, optional_text, text
from ...global_ import global_data
from ...sql import Interval

log = getLogger(__name__)


class DatabaseDisplayer(Displayer):

    @optional_text('Database')
    def _read_date(self, s, date):
        ids = set(d.interval_id for d in global_data().sys.get_dirty_intervals())
        total = s.query(count(Interval.id)).filter(Interval.id.in_(ids)).count()
        today = s.query(count(Interval.id)). \
            filter(Interval.id.in_(ids),
                   Interval.start <= date,
                   Interval.finish > date).scalar()
        if not total:
            yield text('No dirty statistics')
        else:
            if today:
                yield value('Dirty statistics today', today)
            yield value('Total dirty statistics', total)
            yield text('Recalculate statistics')

    def _read_schedule(self, s, date, schedule):
        # todo - include?
        return
        yield




from ..config import Config, WALK, SWIM, RUN, BIKE
from ..database import add_diary_topic, add_child_diary_topic, add_diary_topic_field, add_constant
from ..power import add_simple_power_estimate, add_kit_power_estimate, add_kit_power_model
from ...commands.args import DEFAULT, base_system_path, PERMANENT
from ...diary.model import TYPE, EDIT
from ...lib import to_time, time_to_local_date
from ...msil2a.download import MSIL2A_DIR_CNAME
from ...names import Sports
from ...pipeline.read.activity import ActivityReader
from ...sql import StatisticJournalType, StatisticName, DiaryTopic, DiaryTopicJournal
from ...sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ...sql.utils import add


def acooke(s, base):
    '''
## acooke

This extends the default configuration with:

* Diary entries that I need
* Additional activity groups selected on the kit used
* Power estimates

Unlikely to be useful to others, but works as an example of how you can extend the code yourself.
    '''
    ACooke(base).load(s)


ROAD = 'Road'
MTB = 'MTB'


class ACooke(Config):

    def _load_diary_topics(self, s, c):
        super()._load_diary_topics(s, c)

        # add more diary entries specific to my needs

        injuries = add_diary_topic(s, 'Injuries', c)

        ms = add_child_diary_topic(s, injuries, 'Multiple Sclerosis', c)
        add_diary_topic_field(s, ms, 'MS Notes', c, StatisticJournalType.TEXT, model={TYPE: EDIT})
        add_child_diary_topic(s, ms, 'Betaferon', c,
                              schedule='2018-08-08/2d[1]')  # reminder to take meds on alternate days

        leg = add_child_diary_topic(s, injuries, 'Broken Femur LHS', c,
                                    schedule='d2018-03-11-2020-03-01')
        add_diary_topic_field(s, leg, 'Leg Notes', c, StatisticJournalType.TEXT, model={TYPE: EDIT})
        add_child_diary_topic(s, leg, 'Learn to manage tendon pain', c)  # aims added as child topics
        add_child_diary_topic(s, leg, 'Maintain fitness', c)
        add_child_diary_topic(s, leg, 'Visit UK', c, schedule='-2018-08-11')

    def _load_specific_activity_groups(self, s):
        super()._load_specific_activity_groups(s)

        # additional activity groups for different cycling activities

        self._load_activity_group(s, ROAD, 'Road cycling activities')
        self._load_activity_group(s, MTB, 'MTB cycling activities')

    def _sport_to_activity(self):

        # map the additional groups above based on kit use
        # (cotic and bowman are kit items added via kit commands)

        return {Sports.SPORT_CYCLING: {
                    ActivityReader.KIT: {
                        'cotic': MTB,
                        'bowman': ROAD,
                    },
                    DEFAULT: BIKE,
                },
                Sports.SPORT_RUNNING: RUN,
                Sports.SPORT_SWIMMING: SWIM,
                Sports.SPORT_WALKING: WALK}

    def _load_power_statistics(self, s, c, simple=False):
        # add power estimates for the two bikes
        # (note that this comes after standard stats, but before summary, achievements, etc).
        if simple:
            for activity_group in (MTB, ROAD):
                activity_group = self._activity_groups[activity_group]
                add_simple_power_estimate(s, c, activity_group, 0.42, 0.0055, 12, 65)
        else:
            add_kit_power_estimate(s, c, (MTB, ROAD))
            for kit, activity_group, cda, crr, bike_weight in (('cotic', MTB, 0.42, 0.0055, 12),
                                                               ('bowman', ROAD, 0.42, 0.0055, 8)):
                add_kit_power_model(s, kit, self._activity_groups[activity_group], cda, crr, bike_weight)

    def _load_constants(self, s):
        super()._load_constants(s)
        add_constant(s, MSIL2A_DIR_CNAME,
                     base_system_path(self._base, version=PERMANENT, subdir='msil2a', create=False),
                     description='''
Directory containing Sentinel 2A imaging data (see https://scihub.copernicus.eu/dhus/#/home)

This can be used to generate background images for plots.
I used the data in an experiment to generate 3D plots, but it wasn't very successful.
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)

    def _post(self, s):
        super()._post(s)
        # set a default weight for early power calculations
        weight = s.query(StatisticName).filter(StatisticName.name == 'Weight', StatisticName.owner == DiaryTopic).one()
        diary = add(s, DiaryTopicJournal(date=time_to_local_date(to_time(0.0))))
        add(s, STATISTIC_JOURNAL_CLASSES[weight.statistic_journal_type](
            value=65.0, time=0.0, statistic_name=weight, source=diary))




from logging import getLogger

from .model import text, value
from ..lib import to_date
from ..lib.date import YMD
from ..pipeline.calculate.summary import SummaryCalculator
from ..pipeline.display.utils import Displayer
from ..sql import StatisticJournal, Pipeline, PipelineType, StatisticName

log = getLogger(__name__)


def read_date(s, date):
    yield text(date.strftime('%Y-%m-%d - %A'), tag='title')
    yield from read_pipeline(s, date)


def read_schedule(s, schedule, date):
    yield text(date.strftime(YMD) + ' - Summary for %s' % schedule.describe(), tag='title')
    yield from read_pipeline(s, date, schedule=schedule)


def summary_column(s, schedule, start, name):
    journals = StatisticJournal.at_interval(s, start, schedule, SummaryCalculator, name, SummaryCalculator)
    for named, journal in enumerate(journal for journal in journals if journal.value != 0):
        summary, period, name = SummaryCalculator.parse_title(journal.statistic_name.title)
        if not named:
            yield text(name)
        yield value(summary, journal.value, units=journal.statistic_name.units)


def read_pipeline(session, date, schedule=None):
    '''
    schedule only sent for summary views.
    '''
    date = to_date(date)   # why is this needed?
    for pipeline in Pipeline.all(session, PipelineType.DISPLAY):
        log.info(f'Building {pipeline.cls} ({pipeline.args}, {pipeline.kargs})')
        instance = pipeline.cls(*pipeline.args, **pipeline.kargs)
        if isinstance(instance, Displayer):  # why is this needed?
            data = list(instance.read(session, date, schedule=schedule))
            if data:
                yield data


def interval_column(s, interval, name, owner):
    statistic_journals = s.query(StatisticJournal). \
        join(StatisticName). \
        filter(StatisticJournal.source == interval,
               StatisticName.owner == owner,
               StatisticName.name.like('%' + name)).all()
    for named, statistic_journal in enumerate(statistic_journal
                                              for statistic_journal in statistic_journals
                                              if statistic_journal.value != 0):
        summary, period, name = SummaryCalculator.parse_title(statistic_journal.statistic_name.title)
        if not named:
            yield text(name)
        yield value(summary, statistic_journal.value, units=statistic_journal.statistic_name.units)



from logging import getLogger

from .database import add_statistics, add_enum_constant
from ..names import Titles, Names
from ..pipeline.calculate.impulse import HRImpulse, ImpulseCalculator
from ..pipeline.calculate.response import Response, ResponseCalculator
from ..pipeline.read.segment import SegmentReader
from ..sql import ActivityGroup
from ..sql.types import short_cls

'''
individual impulse calculations are each configured by a separate constant and associated with a 
separate pipeline instance that does the work.  they are grouped by 'prefix' and specific to an
activity group.  so within a prefix there should be no more than one impulse for each group.

the pipeline class generates impulses named by the prefix + the constant, associated with the 
activity group.  in addition, to group ALL, the same value is written with name prefix + 'hr_impulse_10'.
this allows for easy plotting (a prefix of 'default' is assumed to exist).

for each named impulse there are two statistics - one in the given activity group and one in ALL.

the response calculator combines impulses from all statistics with a given prefix using the data in ALL.
for efficiency a single pipeline instance s used for all responses with a given prefix (so the 
pipeline config is a list of constant names rather than the single name used for impulses).

as with impulses, the responses are named by prefix and constant.   
'''

log = getLogger(__name__)


def add_impulse(s, c, activity_group, gamma=2, zero=1, one=6, max_secs=60,
                title=Titles.HR_IMPULSE_10, prefix=Names.DEFAULT):
    '''
    Add configuration for a fitness/fatigue impulse model based on HR zones.
    '''
    if isinstance(activity_group, ActivityGroup):
        activity_group = activity_group.name
    if not activity_group:
        raise Exception(f'Impulse must be defined for a specific group')
    constant = add_enum_constant(s, title, HRImpulse,
                                 {'title': title, 'gamma': 2.0, 'zero': 1, 'one': 6, 'max_secs': 60},
                                 single=True, activity_group=activity_group, description='''
Data needed to calculate the FF-model impulse from heart rate zones.
* Gamma is an exponent used to weight the relative importance of hard and easy efforts and should be around 1 (typical range 0.5 to 2).
* Zero defines the lowest heart zone used (so a value of 2 means that a HR zone of 2.0 or less contributes zero impulse).
* One defines the range of heart rate zones used (so a value of 6 means that a HR zone of 6.0 gives a unit impulse per time interval).
* Max_secs is the maximum gap between heart rate measurements (data that exceed this give no impulse).
Once the impulse is calculated it is summed with a decay to find fitness and fatigue
(see Fitness and Fatigue constants). 
''')
    add_statistics(s, ImpulseCalculator, c, owner_in=short_cls(SegmentReader),
                   impulse_constant=constant.name, prefix=prefix,
                   activity_group=activity_group)


def add_responses(s, c, responses=((42, 1, 1, Titles.FITNESS_D % 42, 'fitness'),
                                   (7, 1, 5, Titles.FATIGUE_D % 7, 'fatigue')),
                  owner_in=ImpulseCalculator, prefix=Names.DEFAULT):
    '''
    Add configuration for a fitness/fatigue response model based on pre-calculated impulses.
    '''
    constants = [add_enum_constant(s, title, Response,
                                   {'src_owner': short_cls(ImpulseCalculator),
                                    'title': title, 'tau_days': days, 'start': start, 'scale': scale},
                                   single=True, description=f'''
Data needed to calculate the FF-model {label} for {days} days.
* Src_owner is the process that generated the input data (ImpulseCalculator calculates the HR impulses).
* Title is the statistic title for display the results.
* Tau_days is the time period (in days) over which the fitness decays.
* Start is the initial fitness value.
* Scale is an arbitrary scale factor (typically used so that fitness and fatigue have comparable values).
''') for (days, start, scale, title, label) in responses]
    add_statistics(s, ResponseCalculator, c,
                   owner_in=short_cls(owner_in), prefix=prefix,
                   response_constants=[constant.name for constant in constants])



from logging import getLogger
from os.path import splitext, basename

from pygeotile.point import Point
from sqlalchemy.sql.functions import count

from .utils import AbortImportButMarkScanned, MultiProcFitReader
from ... import FatalException
from ...commands.args import mm, FORCE, DEFAULT, no, READ
from ...diary.model import TYPE, EDIT
from ...fit.format.records import fix_degrees, merge_duplicates, no_bad_values
from ...fit.profile.profile import read_fit
from ...lib.date import to_time, time_to_local_time
from ...lib.io import split_fit_path
from ...names import N, T, Units, Sports, Summaries as S
from ...sql.database import Timestamp, StatisticJournalText
from ...sql.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...sql.tables.statistic import StatisticJournalFloat, STATISTIC_JOURNAL_CLASSES, StatisticName, \
    StatisticJournalType, StatisticJournal
from ...sql.tables.topic import ActivityTopicField, ActivityTopic, ActivityTopicJournal
from ...sql.utils import add
from ...srtm.bilinear import bilinear_elevation_from_constant

log = getLogger(__name__)


# duplicate data in
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24715592701_tap-sync-18690-cc1dd93225119215a1ea87c584a974ce.fit
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24718989709_tap-sync-18690-effaaaffdd06b9419991471bd92d53d5.fit

class ActivityReader(MultiProcFitReader):

    KIT = 'kit'

    def __init__(self, *args, define=None, sport_to_activity=None, record_to_db=None, **kargs):
        from ...commands.read import ACTIVITY
        self.define = define if define else {}
        self.sport_to_activity = self._assert('sport_to_activity', sport_to_activity)
        self.record_to_db = [(field, title, units, STATISTIC_JOURNAL_CLASSES[type])
                             for field, (title, units, type)
                             in self._assert('record_to_db', record_to_db).items()]
        self.add_elevation = not any(title == T.ELEVATION for (field, title, units, type) in self.record_to_db)
        self.__ajournal = None  # save for coverage
        super().__init__(*args, sub_dir=ACTIVITY, **kargs)

    def _base_command(self):
        force = mm(FORCE) if self.force else ''
        return f'{READ} {force}'

    def _startup(self, s):
        super()._startup(s)
        self.__oracle = bilinear_elevation_from_constant(s)

    def _build_define(self, path):
        define = dict(self.define)
        _, kit = split_fit_path(path)
        if kit:
            if ActivityReader.KIT in define and define[ActivityReader.KIT] != kit:
                log.warning(f'Ignoring {kit} from filename since value given '
                            f'({define[ActivityReader.KIT]}) on command line')
            else:
                define[ActivityReader.KIT] = kit
        else:
            log.debug(f'No {ActivityReader.KIT} in {path}')
        return define

    def _read_data(self, s, file_scan):
        log.info('Reading activity data from %s' % file_scan)
        records = self.parse_records(read_fit(file_scan.path))
        define = self._build_define(file_scan.path)
        ajournal, activity_group, first_timestamp = self._create_activity(s, file_scan, define, records)
        return ajournal, (ajournal, activity_group, first_timestamp, file_scan, define, records)

    @staticmethod
    def parse_records(data):
        return ActivityReader.read_fit_file(data, merge_duplicates, fix_degrees, no_bad_values)

    @staticmethod
    def read_sport(path, records):
        try:
            return ActivityReader._first(path, records, 'sport').value.sport.lower()
        except AbortImportButMarkScanned:
            # alternative for some garmin devices (florian)
            return ActivityReader._first(path, records, 'session').value.sport.lower()

    @staticmethod
    def read_first_timestamp(path, records):
        return ActivityReader._first(path, records, 'event', 'record').value.timestamp

    @staticmethod
    def read_last_timestamp(path, records):
        return ActivityReader._last(path, records, 'event', 'record').value.timestamp

    def _create_activity(self, s, file_scan, define, records):
        first_timestamp = self.read_first_timestamp(file_scan.path, records)
        last_timestamp = self.read_last_timestamp(file_scan.path, records)
        log.debug(f'Time range: {first_timestamp.timestamp()} - {last_timestamp.timestamp()}')
        sport = self.read_sport(file_scan.path, records)
        activity_group = self._activity_group(s, file_scan.path, sport, self.sport_to_activity, define)
        log.info(f'{activity_group} from {sport} / {define}')
        if self.force:
            self._delete_journals(s, activity_group, first_timestamp, last_timestamp, file_scan)
        else:
            self._check_journals(s, activity_group, first_timestamp, last_timestamp, file_scan)
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          file_hash_id=file_scan.file_hash_id))
        return ajournal, activity_group, first_timestamp

    def _activity_group(self, s, path, sport, lookup, define):
        log.debug(f'Current lookup: {lookup}; sport: {sport}; define: {define}')
        if isinstance(lookup, str):
            return self._lookup_activity_group(s, lookup)
        if sport in lookup:
            return self._activity_group(s, path, sport, lookup[sport], define)
        for key, values in define.items():
            for value in values.split(','):
                if key in lookup and value in lookup[key]:
                    return self._activity_group(s, path, sport, lookup[key][value], define)
        if DEFAULT in lookup:
            return self._activity_group(s, path, sport, lookup[DEFAULT], define)
        log.warning('Unrecognised sport: "%s" in %s' % (sport, path))
        if sport in (Sports.SPORT_GENERIC,):
            raise Exception(f'Ignoring {sport} entry')
        else:
            raise FatalException(f'There is no group configured for {sport} entries in the FIT file.')

    def _lookup_activity_group(self, s, name):
        activity_group = ActivityGroup.from_name(s, name)
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                log.info('Available activity group:')
                for activity_group in activities:
                    log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _overlapping_journals(self, s, what, activity_group, first_timestamp, last_timestamp):
        return s.query(what). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.start < last_timestamp,
                       ActivityJournal.finish >= first_timestamp)

    def _file_hash_journals(self, s, what, file_scan):
        return s.query(what). \
                filter(ActivityJournal.file_hash == file_scan.file_hash)

    def _delete_query(self, s, query):
        for journal in query.all():
            Timestamp.clear(s, owner=self.owner_out, source=journal)
            log.debug(f'Deleting {journal}')
            s.delete(journal)

    def _delete_journals(self, s, activity_group, first_timestamp, last_timestamp, file_scan):
        log.debug('Deleting overlapping journals')
        self._delete_query(s, self._overlapping_journals(s, ActivityJournal, activity_group,
                                                         first_timestamp, last_timestamp))
        log.debug('Deleting journals from same file')
        self._delete_query(s, self._file_hash_journals(s, ActivityJournal, file_scan))
        s.commit()

    def _check_journals(self, s, activity_group, first_timestamp, last_timestamp, file_scan):
        log.debug('Checking for overlapping journals')
        if self._overlapping_journals(s, count(ActivityJournal.id), activity_group,
                                      first_timestamp, last_timestamp).scalar():
            log.warning(f'Overlapping activities for {first_timestamp} - {last_timestamp}')
            raise AbortImportButMarkScanned()
        log.debug('Checking for previous file')
        if self._file_hash_journals(s, count(ActivityJournal.id), file_scan).scalar():
            log.warning(f'Repeated scan of file {file_scan.path}')
            raise Exception(file_scan.path)  # this should not happen since hash is checked before import

    def _load_define(self, s, define, ajournal):
        for name, value in define.items():
            log.debug(f'Setting {name} = {value}')
            if name == ActivityReader.KIT:
                description = 'Kit used in activity.'
            else:
                description = 'Attribute defined on reading activity.'
            StatisticJournalText.add(s, name, None, None, self.owner_out,
                                     ajournal, value, ajournal.start, description=description)

    @staticmethod
    def _save_name(s, ajournal, file_scan):
        from ...config.database import add_activity_topic_field
        log.debug('Saving name')
        # first, do we have the 'Name' field defined?
        # this should be triggered at most once per group if it was not already defined
        root = s.query(ActivityTopic). \
            filter(ActivityTopic.title == ActivityTopic.ROOT,
                   ActivityTopic.activity_group == ajournal.activity_group).one_or_none()
        if not root:
            root = add(s, ActivityTopic(name=ActivityTopic.ROOT, description=ActivityTopic.ROOT_DESCRIPTION,
                                        activity_group=ajournal.activity_group))
        if not s.query(ActivityTopicField). \
                join(StatisticName). \
                filter(StatisticName.name == N.NAME,
                       StatisticName.owner == ActivityTopic,
                       ActivityTopicField.activity_topic == root).one_or_none():
           add_activity_topic_field(s, root, T.NAME, -10, StatisticJournalType.TEXT,
                                     ajournal.activity_group, model={TYPE: EDIT},
                                     description=ActivityTopicField.NAME_DESCRIPTION)
        # second, do we already have a journal for this file, or do we need to add one?
        source = s.query(ActivityTopicJournal). \
            filter(ActivityTopicJournal.file_hash_id == file_scan.file_hash_id). \
            one_or_none()
        if not source:
            source = add(s, ActivityTopicJournal(file_hash_id=file_scan.file_hash_id,
                                                 activity_group=ajournal.activity_group))
        # and third, does that journal contain a name?
        if not s.query(StatisticJournal). \
                join(StatisticName). \
                filter(StatisticJournal.source == source,
                       StatisticName.owner == ActivityTopic,
                       StatisticName.name == N.NAME).one_or_none():
            value = splitext(basename(file_scan.path))[0]
            StatisticJournalText.add(s, N.NAME, None, None, ActivityTopic,
                                     source, value, ajournal.start)

    def _check_overlap(self, s, start, finish, ajournal):
        overlap = s.query(ActivityJournal). \
            filter(ActivityJournal.finish >= start,
                   ActivityJournal.start <= finish,
                   ActivityJournal.id != ajournal.id).first()
        if overlap:
            def fmt(ajournal):
                return f'{ajournal} {ajournal.file_hash.file_scan.path}'
            log.error(f'Overlapping activities: {fmt(ajournal)} / {fmt(overlap)}')
            raise Exception(f'Overlapping activities: '
                            f'{time_to_local_time(ajournal.start)} / {time_to_local_time(overlap.start)}')

    def _load_data(self, s, loader, data):

        ajournal, activity_group, first_timestamp, file_scan, define, records = data
        timespan, warned, logged, last_timestamp = None, 0, 0, to_time(0.0)

        log.debug(f'Loading {self.record_to_db}')

        def is_event(record, *types):
            event = False
            if record.name == 'event':
                event = record.value.event == 'timer' and record.value.event_type in types
                if event: log.debug(f'{types} at {record.timestamp}')
            return event

        have_timespan = any(is_event(record, 'start') for record in records)
        only_records = list(filter(lambda x: x.name == 'record', records))
        final_timestamp = only_records[-1].timestamp

        self._check_overlap(s, first_timestamp, final_timestamp, ajournal)
        self._load_define(s, define, ajournal)
        self._save_name(s, ajournal, file_scan)
        self.__ajournal = ajournal

        if not have_timespan:
            first_timestamp = only_records[0].timestamp
            log.warning('Experimental handling of data without timespans')
            timespan = add(s, ActivityTimespan(activity_journal=ajournal, start=first_timestamp))

        for record in records:

            if have_timespan and is_event(record, 'start'):
                if timespan:
                    log.warning('Ignoring start with no corresponding stop (possible lost data?)')
                else:
                    timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                       start=record.value.timestamp,
                                                       finish=record.value.timestamp))

            elif record.name == 'record':
                if record.value.timestamp > last_timestamp:
                    lat, lon, timestamp = None, None, record.value.timestamp
                    # customizable loader
                    for field, title, units, type in self.record_to_db:
                        value = record.data.get(field, None)
                        if logged < 3:
                            log.debug(f'{title} = {value}')
                        if value is not None:
                            value = value[0][0]
                            if units == Units.KM:  # internally everything uses M
                                value /= 1000
                            loader.add(title, units, None, ajournal, value, timestamp, type,
                                       description=f'The value of field {field} in the FIT record.')
                            if title == T.LATITUDE:
                                lat = value
                            elif title == T.LONGITUDE:
                                lon = value
                    logged += 1
                    # values derived from lat/lon
                    if lat is not None and lon is not None:
                        x, y = Point.from_latitude_longitude(lat, lon).meters
                        loader.add(T.SPHERICAL_MERCATOR_X, Units.M, None, ajournal, x, timestamp,
                                   StatisticJournalFloat, description='The WGS84 X coordinate')
                        loader.add(T.SPHERICAL_MERCATOR_Y, Units.M, None, ajournal, y, timestamp,
                                   StatisticJournalFloat, description='The WGS84 Y coordinate')
                        if self.add_elevation:
                            elevation = self.__oracle.elevation(lat, lon)
                            if elevation:
                                loader.add(T.RAW_ELEVATION, Units.M, None, ajournal, elevation,
                                           timestamp, StatisticJournalFloat,
                                           description='The elevation from SRTM1 at this location')
                else:
                    log.warning('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                (file_scan.path, record.value.timestamp))
                last_timestamp = record.value.timestamp
                if not have_timespan:
                    ajournal.finish = record.value.timestamp

            elif have_timespan and is_event(record, 'stop_all', 'stop'):
                if timespan:
                    timespan.finish = record.value.timestamp
                    ajournal.finish = record.value.timestamp
                    timespan = None
                else:
                    log.debug('Ignoring stop with no corresponding start (possible lost data?)')

        if timespan:
            log.warning('Cleaning up dangling timespan')
            timespan.finish = final_timestamp

    def _read(self, s, path):
        loader = super()._read(s, path)
        for title, percent in loader.coverage_percentages():
            StatisticJournalFloat.add(s, T._cov(title), Units.PC, S.join(S.MIN, S.AVG), self.owner_out,
                                      self.__ajournal, percent, self.__ajournal.start,
                                      description=f'Coverage (% of FIT records with data) for {title}.')
        s.commit()




from logging import getLogger
from random import choice

from sqlalchemy import func, inspect, and_, select
from sqlalchemy.sql.functions import coalesce

from .utils import MultiProcCalculator, IntervalCalculatorMixin
from ...data.frame import _tables
from ...global_ import global_data
from ...names import Summaries as S
from ...lib.date import local_date_to_time
from ...sql.tables.source import Interval, Source
from ...sql.tables.statistic import StatisticJournal, StatisticName, StatisticMeasure, StatisticJournalInteger, \
    StatisticJournalFloat, StatisticJournalText, TYPE_TO_JOURNAL_CLASS, STATISTIC_JOURNAL_CLASSES

log = getLogger(__name__)


def fuzz(n, q):
    # n is number of points, q is quartile (0-4).
    # return index for quartile that is unbiased when location not exact
    i = (n-1) * q / 4
    if i != int(i):
        i = int(i) + choice([0, 1])  # if we're between two points, pick either
    return int(i)


class SummaryCalculator(IntervalCalculatorMixin, MultiProcCalculator):

    # todo - this should have a worker per activity group

    def __init__(self, *args, grouped=True, **kargs):
        super().__init__(*args, grouped=grouped, **kargs)

    def _startup(self, s):
        Interval.clean(s)

    def _read_data(self, s, interval):
        # here, data is only statistics names, because calculation also involves loading data
        start, finish = local_date_to_time(interval.start), local_date_to_time(interval.finish)
        statistics_with_data_and_summary = s.query(StatisticName.id). \
            join(StatisticJournal). \
            filter(StatisticJournal.time >= start,
                   StatisticJournal.time < finish,
                   StatisticName.summary != None)
        # avoid duplicates
        return s.query(StatisticName). \
            filter(StatisticName.id.in_(statistics_with_data_and_summary)).all()

    def _calculate_results(self, s, interval, data, loader):
        log.debug('Calculating summaries')
        start, finish = local_date_to_time(interval.start), local_date_to_time(interval.finish)
        measures = []
        for statistic_name in data:
            summaries = statistic_name.summaries
            for summary in summaries:
                value, units = self._calculate_value(s, statistic_name, summary, S.MIN in summaries,
                                                     start, finish, interval, measures)
                if value is not None:
                    title = self.fmt_title(statistic_name.title, summary, self.schedule)
                    # we need to infer the type
                    if summary in (S.MAX, S.MIN, S.SUM):
                        new_type = TYPE_TO_JOURNAL_CLASS[type(value)]
                    elif summary in (S.AVG,):
                        new_type = StatisticJournalFloat
                    else:
                        new_type = StatisticJournalInteger
                    loader.add(title, units, None, interval, value, start, new_type,
                               description=self._describe(statistic_name, summary, interval))
        # add and commit these here - what else can we do?
        log.debug(f'Adding {len(measures)} measures')
        for measure in measures:
            s.add(measure)
        s.commit()

    def _calculate_value(self, s, statistic_name, summary, order_asc, start_time, finish_time, interval, measures):

        t = _tables()
        sjx = inspect(STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]).local_table
        units = statistic_name.units
        activity_group_id = interval.activity_group.id if interval.activity_group else None

        if summary == S.MAX:
            result = func.max(sjx.c.value)
        elif summary == S.MIN:
            result = func.min(sjx.c.value)
        elif summary == S.SUM:
            result = func.sum(sjx.c.value)
        elif summary == S.CNT:
            result = func.count(sjx.c.value)
            units = None
        elif summary == S.AVG:
            result = func.avg(sjx.c.value)
        elif summary == S.MSR:
            self._calculate_measures(s, statistic_name, order_asc, start_time, finish_time, interval, measures)
            return None, None
        else:
            raise Exception('Bad summary: %s' % summary)

        stmt = select([result]). \
            select_from(sjx).select_from(t.sj).select_from(t.src). \
            where(and_(t.sj.c.id == sjx.c.id,
                       t.sj.c.statistic_name_id == statistic_name.id,
                       t.sj.c.time >= start_time,
                       t.sj.c.time < finish_time,
                       t.sj.c.source_id == t.src.c.id,
                       t.src.c.activity_group_id == activity_group_id))

        return next(s.connection().execute(stmt))[0], units

    def _describe(self, statistic_name, summary, interval):
        adjective = {S.MAX: 'highest', S.MIN: 'lowest', S.SUM: 'total', S.CNT: 'number of', S.AVG: 'average'}[summary]
        period = interval.schedule.describe().lower()
        if period == 'all':
            period = period + ' time'
        else:
            period = 'one ' + period
        return f'The {adjective} {statistic_name.title} over {period}.'

    def _calculate_measures(self, s, statistic_name, order_asc, start_time, finish_time, interval, measures):
        data = sorted([x for x in
                       s.query(StatisticJournal).
                      join(Source).
                      filter(StatisticJournal.statistic_name == statistic_name,
                             StatisticJournal.time >= start_time,
                             StatisticJournal.time < finish_time,
                             Source.activity_group == interval.activity_group).all()
                       if x is not None and x.value is not None],
                      key=lambda x: x.value, reverse=not order_asc)
        n, local_measures = len(data), []
        for rank, journal in enumerate(data, start=1):
            if n > 1:
                percentile = (n - rank) / (n - 1) * 100
            else:
                percentile = 100
            measure = StatisticMeasure(statistic_journal=journal, source=interval, rank=rank, percentile=percentile)
            local_measures.append(measure)
            measures.append(measure)
        if n > 8:  # avoid overlap in fuzzing (and also, plot individual points in this case)
            for q in range(5):
                local_measures[fuzz(n, q)].quartile = q
        log.debug('Ranked %s' % statistic_name)

    @classmethod
    def parse_title(cls, name):
        left, right = name.split(' ', 1)
        summary, period = left.split('/')
        return summary, period, right

    @classmethod
    def fmt_title(cls, name, summary, schedule):
        title = summary.capitalize()
        return '%s/%s %s' % (title, schedule.describe(), name)



from logging import getLogger

from sqlalchemy import func, distinct

from ..json import JsonResponse
from ...commands.search import text_search
from ...data import constrained_sources
from ...data.constraint import activity_conversion
from ...diary.model import VALUE, UNITS, DB
from ...lib import time_to_local_time
from ...lib.utils import parse_bool
from ...names import N
from ...pipeline.calculate import ActivityCalculator
from ...sql import StatisticName, ActivityGroup, StatisticJournal, Source, ActivityTopicField, ActivityTopic
from ...sql.tables.source import SourceType

log = getLogger(__name__)

NAME = 'name'
DESCRIPTION = 'description'
GROUPS = 'groups'

RESULTS = 'results'
ERROR = 'error'


class Search:

    @staticmethod
    def query_activity(request, s, query):
        try:
            advanced = parse_bool(request.args.get('advanced', 'false'), default=None)
            log.info(f'{"advanced " if advanced else ""}query: {query}')
            return JsonResponse({RESULTS: search(s, query, advanced)})
        except Exception as e:
            log.warning(e)
            return JsonResponse({ERROR: str(e)})

    @staticmethod
    def read_activity_terms(request, s):

        def format(row):
            name, description, units = row
            return {NAME: name, DESCRIPTION: description, UNITS: units}

        used = s.query(distinct(StatisticJournal.statistic_name_id)). \
            join(Source, Source.id == StatisticJournal.source_id). \
            filter(Source.type.in_([SourceType.ACTIVITY_TOPIC, SourceType.ACTIVITY]))
        q = s.query(StatisticName.name, StatisticName.description, StatisticName.units). \
            filter(StatisticName.id.in_(used))
        log.debug(q)
        return JsonResponse([format(row) for row in q.all()])


def search(s, query, advanced):
    if advanced:
        activities = constrained_sources(s, query, conversion=activity_conversion)
    else:
        activities = text_search(s, query)
    return [expand_activity(s, activity) for activity in sorted(activities, key=lambda x: x.start)]


def expand_activity(s, activity_journal):

    topic_journal = activity_journal.get_activity_topic_journal(s)

    def format(statistic):
        if statistic:
            return {VALUE: statistic.value, UNITS: statistic.statistic_name.units}
        else:
            return None

    def value(statistic):
        if statistic:
            return statistic.value
        else:
            return None

    return {DB: activity_journal.id,
            'name': {VALUE: value(StatisticJournal.for_source(s, topic_journal.id, N.NAME,
                                                              ActivityTopic, topic_journal.activity_group)),
                     UNITS: None},
            'group': {VALUE: activity_journal.activity_group.name, UNITS: None},
            'start': {VALUE: time_to_local_time(activity_journal.start), UNITS: 'date'},
            'time': format(StatisticJournal.for_source(s, activity_journal.id, N.ACTIVE_TIME,
                                                       ActivityCalculator, activity_journal.activity_group)),
            'distance': format(StatisticJournal.for_source(s, activity_journal.id, N.ACTIVE_DISTANCE,
                                                           ActivityCalculator, activity_journal.activity_group))}



from collections import namedtuple
from logging import getLogger

from sqlalchemy.orm.exc import NoResultFound

from ..commands.args import DB_VERSION
from ..commands.database import database_really_exists
from ..lib import format_date, time_to_local_date, to_time
from ..sql import StatisticJournal, StatisticName, StatisticJournalType
from ..sql.database import sqlite_uri, postgresql_uri
from ..sql.tables.statistic import STATISTIC_JOURNAL_CLASSES
from ..sql.types import short_cls
from ..sql.utils import add

log = getLogger(__name__)


def journal_imported(record, new, cls, name, allow_time_zero=False):
    # true if already installed
    with new.session_context() as new_s:
        q = new_s.query(StatisticJournal). \
            join(cls). \
            filter(StatisticJournal.source_id == cls.id)
        if allow_time_zero:
            q = q.filter(StatisticJournal.time > 0.0)
        if q.count():
            record.warning(f'{name} topic entries already exist - old data must be imported first')
            return True
    return False


def match_statistic_name(record, old_statistic_name, new_s, owner):
    try:
        log.debug(f'Trying to find new statistic_name for {old_statistic_name} ({owner})')
        new_statistic_name = new_s.query(StatisticName). \
            filter(StatisticName.name == old_statistic_name.name,
                   StatisticName.owner == owner,
                   StatisticName.statistic_journal_type == old_statistic_name.statistic_journal_type).one()
        log.debug(f'Found new statistic_name {new_statistic_name}')
        return new_statistic_name
    except NoResultFound:
        record.raise_(f'No new equivalent to statistic {old_statistic_name.name} '
                      f'({StatisticJournalType(old_statistic_name.statistic_journal_type).name}) '
                      f'for {short_cls(owner)}')


def copy_statistic_journal(record, old_s, old, old_statistic_name, old_statistic_journal,
                           new_s, new_statistic_name, source, name=None):
    name = name if name else new_statistic_name.name
    old_journals = {StatisticJournalType.INTEGER.value: old.meta.tables['statistic_journal_integer'],
                    StatisticJournalType.FLOAT.value: old.meta.tables['statistic_journal_float'],
                    StatisticJournalType.TEXT.value: old.meta.tables['statistic_journal_text']}
    old_journal = old_journals[old_statistic_name.statistic_journal_type]
    old_value = old_s.query(old_journal).filter(old_journal.c.id == old_statistic_journal.id).one()
    log.debug(f'Resolved old statistic_journal {old_value}')
    new_journal = STATISTIC_JOURNAL_CLASSES[StatisticJournalType(new_statistic_name.statistic_journal_type)]
    previous = new_s.query(new_journal). \
        filter(new_journal.time == old_statistic_journal.time,
               new_journal.statistic_name == new_statistic_name).one_or_none()
    # drop ugly auto-titles if nicer ones available (bug fix 0-32 to 0-33)
    if previous and new_statistic_name.name == 'name' and \
            new_statistic_name.statistic_journal_type == StatisticJournalType.TEXT and \
            previous.value.startswith('20'):
        record.warning(f'Dropping previous ({previous}) for {name}')
        new_s.delete(previous)
        new_s.commit()
        previous = None
    if previous:
        record.warning(f'Value already exists for {name} ({previous})')
    else:
        new_value = add(new_s,
                        new_journal(value=old_value.value, time=old_statistic_journal.time,
                                    statistic_name=new_statistic_name, source=source))
        new_s.commit()  # avoid logging below if error
        date = format_date(time_to_local_date(to_time(new_value.time)))
        record.info(f'Statistic value {new_value.value} at {date} for {name}')


def any_attr(instance, *names):
    for name in names:
        if hasattr(instance, name):
            return getattr(instance, name)
    raise AttributeError(f'No {names} in {instance} ({type(instance)})')


def clone_with(result, **kargs):
    return namedtuple('clone', result._fields, defaults=result)(**kargs)


def available_versions(data, max_depth=3):
    return [version[1] for version in sorted(find_versions(data, max_depth=max_depth), reverse=True)]


def find_versions(data, max_depth=3):
    current_version = [int(version) for version in DB_VERSION.split('-')]
    yield from find_sqlite_from_base_and_version(data.base, current_version, max_depth=max_depth)
    yield from find_postgresql_from_version(current_version, max_depth=max_depth)


def count_down_version(current_version, restart_minor=0, max_depth=3):
    major, minor = current_version
    minor -= 1  # skip current
    while major >= 0:
        while minor >= 0 and max_depth > 0:
            yield major, minor
            minor -= 1
            max_depth -= 1
        minor = restart_minor
        major -= 1


def find_sqlite_from_base_and_version(base, current_version, max_depth=3):
    for major, minor in count_down_version(current_version, max_depth=max_depth):
        version = f'{major}-{minor}'
        uri = sqlite_uri(base, version=version)
        if database_really_exists(uri):
            yield version, uri


def find_postgresql_from_version(current_version, max_depth=3):
    for major, minor in count_down_version(current_version, max_depth=max_depth):
        version = f'{major}-{minor}'
        uri = postgresql_uri(version=version)
        if database_really_exists(uri):
            yield version, uri



from logging import getLogger

from sqlalchemy import Column, Integer, ForeignKey, create_engine, union
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.sql.functions import count

from tests import LogTestCase


log = getLogger(__name__)
Base = declarative_base()


'''
the idea here is that we have some nodes in a 'directed graph'.  that means that some nodes
are the 'input' to other nodes.  that relationship is described in the connect table, which 
has an entry for each connection.

in addition, each node has a required number of inputs.  the problem, then, is to detect when
this is inconsistent with the actual number of inputs.

we can solve that problem directly with a normal SQL query.  however, when we delete nodes 
that are inconsistent we can introduce more problems with other nodes.  to delete all nodes
and leave a consistent graph we need a recursive query.

this problem may seem a little contrived.  the motivation is tracking the provenance of
calculated data in choochoo - some statistics may be calculated from multiple sources and
we need to detect which statistics become invalid when some sources are deleted.  that
kind of problem can often be solved with 'on delete cascade', but in the case of choochoo
it is complicated by 'table inheritance' (inheritance in the OO model mapped to the 
database).  the end result is that we need to do this 'garbage collection' manually.  
'''


class Node(Base):

    __tablename__ = 'node'

    id = Column(Integer, primary_key=True)
    n_input = Column(Integer, nullable=False)


class Connect(Base):

    __tablename__ = 'connect'

    id = Column(Integer, primary_key=True)
    input_id = Column(Integer, ForeignKey('node.id', ondelete='cascade'), nullable=False)
    input = relationship('Node', foreign_keys=[input_id])
    output_id = Column(Integer, ForeignKey('node.id', ondelete='cascade'), nullable=False)
    output = relationship('Node', foreign_keys=[output_id])


class RecursiveTest(LogTestCase):

    def setUp(self):
        super().setUp()
        self.engine = create_engine('sqlite:///:memory:', echo=True)
        Base.metadata.create_all(self.engine)
        self.session = sessionmaker(bind=self.engine)()

        '''
        construct the following graph:
        
        1 --\
             > 3 --> 5 -\
        2 --/            > 7 --> 8 
               4     6 -/
              
        note that 4 is not connected to anything and that 5 has only one input, when it expects 2
        (not shown in the diagram above, but set below).
        
        this means that node 5 is inconsistent.  removing that would make node 7 inconsistent and
        removing that would make node 8 inconsistent.  so we want a query that finds (and allows us
        to delete) nodes 5, 7 and 8. 
        '''

        for i in range(8):
            self.session.add(Node(n_input=0))
        self.session.add(Connect(input_id=1, output_id=3))
        self.session.add(Connect(input_id=2, output_id=3))
        self.session.query(Node).filter(Node.id == 3).one().n_input = 2
        self.session.add(Connect(input_id=3, output_id=5))
        self.session.query(Node).filter(Node.id == 5).one().n_input = 2  # wrong!  4 is not an input to 5
        self.session.add(Connect(input_id=5, output_id=7))
        self.session.add(Connect(input_id=6, output_id=7))
        self.session.query(Node).filter(Node.id == 7).one().n_input = 2
        self.session.add(Connect(input_id=7, output_id=8))
        self.session.query(Node).filter(Node.id == 8).one().n_input = 1
        self.session.flush()

    def test_non_zero_inputs(self):
        '''
        the most simple query i can think of, which returns the number of inputs for nodes
        with more than one input
        '''
        q_n_inputs = self.session.query(Connect.output_id.label('id'), count(Connect.input_id).label('count')). \
            group_by(Connect.output_id).order_by(Connect.output_id)
        self.assertEqual([(3, 2), (5, 1), (7, 2), (8, 1)], q_n_inputs.all())

    def test_all_inputs(self):
        '''
        we can improve th eabove to include nodes with zero inputs.
        '''
        q_n_inputs = self.session.query(Node.id.label('id'), count(Connect.input_id).label('count')). \
            outerjoin(Connect, Node.id == Connect.output_id). \
            group_by(Node.id).order_by(Node.id)
        print(q_n_inputs)
        self.assertEqual([(1, 0), (2, 0), (3, 2), (4, 0), (5, 1), (6, 0), (7, 2), (8, 1)], q_n_inputs.all())

    def test_missing_input(self):
        '''
        using the query above as a sub-query, we can compare the actual number of inputs with
        what was expected and find nodes with missing inputs.
        '''
        q_n_inputs = self.session.query(Node.id.label('id'), count(Connect.input_id).label('count')). \
            outerjoin(Connect, Node.id == Connect.output_id). \
            group_by(Node.id).order_by(Node.id).subquery()
        q_missing = self.session.query(Node.id.label('id')). \
            join(q_n_inputs, q_n_inputs.c.id == Node.id). \
            filter(Node.n_input != q_n_inputs.c.count).order_by(Node.id)
        print(q_missing)
        self.assertEqual([(5,)], q_missing.all())

    def test_chained_node(self):
        '''
        we can move one step along the chain by looking for nodes whose inputs will be deleted.
        '''
        q_counts = self.session.query(Node.id.label('id'), count(Connect.input_id).label('count')). \
            outerjoin(Connect, Node.id == Connect.output_id). \
            group_by(Node.id).order_by(Node.id).subquery()
        q_missing = self.session.query(Node.id.label('id')). \
            join(q_counts, q_counts.c.id == Node.id). \
            filter(Node.n_input != q_counts.c.count)
        q_chained = self.session.query(Node.id). \
            join(Connect, Node.id == Connect.output_id). \
            filter(Connect.input_id.in_(q_missing))
        q_all = union(q_missing, q_chained)
        print('\nchained node\n%s\n' % q_all.select())
        self.assertEqual([(5,), (7,)],
                         self.session.query(Node.id).filter(Node.id.in_(q_all.select())).order_by(Node.id).all())

    def test_recursive(self):
        '''
        but to get all nodes we need to recurse...
        '''
        q_counts = self.session.query(Node.id.label('id'), count(Connect.input_id).label('count')). \
            outerjoin(Connect, Node.id == Connect.output_id). \
            group_by(Node.id).order_by(Node.id).subquery()
        q_missing = self.session.query(Node.id.label('id')). \
            join(q_counts, q_counts.c.id == Node.id). \
            filter(Node.n_input != q_counts.c.count).cte(recursive=True)
        q_missing = q_missing.union_all(self.session.query(Node.id).
                                        join(Connect, Node.id == Connect.output_id).
                                        join(q_missing, Connect.input_id == q_missing.c.id))
        print('\nrecursive\n%s\n' % q_missing.select())
        self.assertEqual([(5,), (7,), (8,)],
                         self.session.query(Node.id).filter(Node.id.in_(q_missing.select())).order_by(Node.id).all())
        # this part of test assumes this test runs last in case (since nodes are deleted)
        # self.session.query(Node).filter(Node.id.in_(q_missing.select())).delete(synchronize_session=False)
        # self.session.flush()
        # self.assertEqual([(1,), (2,), (3,), (4,), (6,)], self.session.query(Node.id).order_by(Node.id).all())

    def test_bug(self):
        '''
        so why does this work, without 'recursive'?

        EDIT: the 'recursive' is optional in sqlite!  see the very last line at
        https://www.sqlite.org/lang_with.html

        damn.  and all that trouble to make a nice bug report.
        '''
        q_counts = self.session.query(Node.id.label('id'), count(Connect.input_id).label('count')). \
            outerjoin(Connect, Node.id == Connect.output_id). \
            group_by(Node.id).order_by(Node.id).subquery()
        q_missing = self.session.query(Node.id.label('id')). \
            join(q_counts, q_counts.c.id == Node.id). \
            filter(Node.n_input != q_counts.c.count).cte()
        q_missing = q_missing.union_all(self.session.query(Node.id).
                                        join(Connect, Node.id == Connect.output_id).
                                        join(q_missing, Connect.input_id == q_missing.c.id))
        print('\nbug\n%s\n' % q_missing.select())
        self.assertEqual([(5,), (7,), (8,)],
                         self.session.query(Node.id).filter(Node.id.in_(q_missing.select())).order_by(Node.id).all())

    def test_dump(self):
        cnx = self.engine.raw_connection()
        for line in cnx.iterdump():
            print(line)

'''
bug report details:

BEGIN TRANSACTION;
CREATE TABLE connect (
	id INTEGER NOT NULL, 
	input_id INTEGER NOT NULL, 
	output_id INTEGER NOT NULL, 
	PRIMARY KEY (id), 
	FOREIGN KEY(input_id) REFERENCES node (id) ON DELETE cascade, 
	FOREIGN KEY(output_id) REFERENCES node (id) ON DELETE cascade
);
INSERT INTO "connect" VALUES(1,1,3);
INSERT INTO "connect" VALUES(2,2,3);
INSERT INTO "connect" VALUES(3,3,5);
INSERT INTO "connect" VALUES(4,5,7);
INSERT INTO "connect" VALUES(5,6,7);
INSERT INTO "connect" VALUES(6,7,8);
CREATE TABLE node (
	id INTEGER NOT NULL, 
	n_input INTEGER NOT NULL, 
	PRIMARY KEY (id)
);
INSERT INTO "node" VALUES(1,0);
INSERT INTO "node" VALUES(2,0);
INSERT INTO "node" VALUES(3,2);
INSERT INTO "node" VALUES(4,0);
INSERT INTO "node" VALUES(5,2);
INSERT INTO "node" VALUES(6,0);
INSERT INTO "node" VALUES(7,2);
INSERT INTO "node" VALUES(8,1);
COMMIT;

  WITH RECURSIVE anon_1(id) 
    AS (SELECT node.id AS id 
          FROM node 
          JOIN (SELECT node.id AS id, 
                       count(connect.input_id) AS count 
                  FROM node 
                  LEFT OUTER JOIN connect 
                    ON node.id = connect.output_id 
                 GROUP BY node.id ORDER BY node.id) 
            AS anon_2 
            ON anon_2.id = node.id 
         WHERE node.n_input != anon_2.count 
         UNION ALL 
        SELECT node.id AS node_id 
          FROM node 
          JOIN connect 
            ON node.id = connect.output_id 
          JOIN anon_1 
            ON connect.input_id = anon_1.id)
SELECT node.id AS node_id 
  FROM node 
 WHERE node.id 
    IN (SELECT anon_1.id 
          FROM anon_1)
 ORDER BY node.id;

% 5 7 8

  WITH anon_1 
    AS (SELECT node.id AS id 
          FROM node 
          JOIN (SELECT node.id AS id, 
                       count(connect.input_id) AS count 
                  FROM node 
                  LEFT OUTER JOIN connect 
                    ON node.id = connect.output_id 
                 GROUP BY node.id 
                 ORDER BY node.id) 
            AS anon_2 
            ON anon_2.id = node.id 
         WHERE node.n_input != anon_2.count 
         UNION ALL 
        SELECT node.id AS node_id 
          FROM node 
          JOIN connect 
            ON node.id = connect.output_id 
          JOIN anon_1 
            ON connect.input_id = anon_1.id)
SELECT node.id AS node_id 
  FROM node 
 WHERE node.id 
    IN (SELECT anon_1.id 
          FROM anon_1) 
 ORDER BY node.id;
 
% 5 7 8
'''



from logging import getLogger

from .climb import add_climb, CLIMB_CNAME
from .database import add_loader_support, add_activity_group, add_activities, Counter, add_statistics, add_displayer, \
    add_monitor, add_constant, add_diary_topic, add_diary_topic_field, \
    add_activity_topic_field, add_activity_displayer_delegate, add_activity_topic
from .impulse import add_responses, add_impulse
from ..commands.args import base_system_path, DB_VERSION, PERMANENT
from ..commands.garmin import GARMIN_USER, GARMIN_PASSWORD
from ..diary.model import TYPE, EDIT, FLOAT, LO, HI, DP, SCORE
from ..lib.schedule import Schedule
from ..names import N, Titles, Sports, Units, Summaries as S
from ..pipeline.calculate.achievement import AchievementCalculator
from ..pipeline.calculate.activity import ActivityCalculator
from ..pipeline.calculate.elevation import ElevationCalculator
from ..pipeline.calculate.heart_rate import RestHRCalculator
from ..pipeline.calculate.kit import KitCalculator
from ..pipeline.calculate.monitor import MonitorCalculator
from ..pipeline.calculate.nearby import SimilarityCalculator, NearbyCalculator
from ..pipeline.calculate.response import ResponseCalculator
from ..pipeline.calculate.segment import SegmentCalculator
from ..pipeline.calculate.summary import SummaryCalculator
from ..pipeline.display.activity.achievement import AchievementDelegate
from ..pipeline.display.activity.utils import ActivityDisplayer, ActivityDelegate
from ..pipeline.display.activity.jupyter import JupyterDelegate
from ..pipeline.display.activity.nearby import NearbyDelegate
from ..pipeline.display.activity.segment import SegmentDelegate
from ..pipeline.display.database import DatabaseDisplayer
from ..pipeline.display.diary import DiaryDisplayer
from ..pipeline.display.monitor import MonitorDisplayer
from ..pipeline.display.response import ResponseDisplayer
from ..pipeline.read.monitor import MonitorReader
from ..pipeline.read.segment import SegmentReader
from ..sql import DiaryTopicJournal, StatisticJournalType, ActivityTopicField, SystemConstant, ActivityGroup, \
    ActivityTopic
from ..sql.types import short_cls
from ..srtm.file import SRTM1_DIR_CNAME

log = getLogger(__name__)

NOTES = 'Notes'

BROKEN = 'broken'

BIKE = 'Bike'
RUN = 'Run'
SWIM = 'Swim'
WALK = 'Walk'


class Config:
    '''
    A class-based approach so that we can easily modify the config for different profiles.
    '''

    def __init__(self, base):
        self._base = base
        self._activity_groups = {}

    def load(self, s):
        # hopefully you won't need to over-ride this, but instead one of the more specific methods
        self._pre(s)
        add_loader_support(s)  # required by standard statistics calculations
        self._load_specific_activity_groups(s)
        self._load_activities_pipeline(s, Counter())
        self._load_statistics_pipeline(s, Counter())
        self._load_diary_pipeline(s, Counter())
        self._load_monitor_pipeline(s, Counter())
        self._load_constants(s)
        self._load_diary_topics(s, Counter())
        self._load_activity_topics(s, Counter())
        self._post_diary(s)
        self._post(s)

    def _pre(self, s):
        pass

    def _post_diary(self, s):
        pass

    def _post(self, s):
        pass

    def _load_specific_activity_groups(self, s):
        # statistic rankings (best of month etc) are calculated per group
        self._load_activity_group(s, BIKE, 'Cycling activities')
        self._load_activity_group(s, RUN, 'Running activities')
        self._load_activity_group(s, SWIM, 'Swimming activities')
        self._load_activity_group(s, WALK, 'Walking activities')

    def _load_activity_group(self, s, name, description):
        log.debug(f'Loading activity group {name}')
        self._activity_groups[name] = \
            add_activity_group(s, name, len(self._activity_groups), description=description)

    def _sport_to_activity(self):
        # sport_to_activity maps from the FIT sport field to the activity defined above.
        # more complex mapping is supported using kits (see acooke configuration),
        return {Sports.SPORT_CYCLING: BIKE,
                Sports.SPORT_RUNNING: RUN,
                Sports.SPORT_SWIMMING: SWIM,
                Sports.SPORT_WALKING: WALK}

    def _record_to_db(self):
        # the mapping from FIT fields to database entries
        # you really don't want to alter this unless you know what you are doing...
        # todo - should this depend on activity group?
        return {'position_lat': (Titles.LATITUDE, Units.DEG, StatisticJournalType.FLOAT),
                'position_long': (Titles.LONGITUDE, Units.DEG, StatisticJournalType.FLOAT),
                'heart_rate': (Titles.HEART_RATE, Units.BPM, StatisticJournalType.INTEGER),
                'enhanced_speed': (Titles.SPEED, Units.MS, StatisticJournalType.FLOAT),
                'distance': (Titles.DISTANCE, Units.KM, StatisticJournalType.FLOAT),
                'enhanced_altitude': (Titles.ALTITUDE, Units.M, StatisticJournalType.FLOAT),
                'cadence': (Titles.CADENCE, Units.RPM, StatisticJournalType.INTEGER)}

    def _load_activities_pipeline(self, s, c):
        sport_to_activity = self._sport_to_activity()
        record_to_db = self._record_to_db()
        add_activities(s, SegmentReader, c, owner_out=short_cls(SegmentReader),
                       sport_to_activity=sport_to_activity, record_to_db=record_to_db)

    def _ff_parameters(self):
        return ((42, 1, 1, Titles.FITNESS_D % 42, 'fitness'),
                (7, 1, 5, Titles.FATIGUE_D % 7, 'fatigue'))

    def _load_power_statistics(self, s, c):
        # after elevation and before ff etc
        # defaults is none because configuration is complex
        pass

    def _load_ff_statistics(self, s, c, default_fthr=154):
        for activity_group in self._activity_groups.values():
            add_impulse(s, c, activity_group)
            # we need a value here for various UI reasons.  might as well use my own value...
            add_constant(s, Titles.FTHR, default_fthr,
                         description=f'''
Heart rate (in bpm) at functional threshold.

Your FTHR is the highest sustained heart rate you can maintain for long periods (an hour).
It is used to calculate how hard you are working (the Impulse) and, from that, 
your FF-model parameters (fitness and fatigue).
''',
                         activity_group=activity_group, units=Units.BPM,
                         statistic_journal_type=StatisticJournalType.INTEGER)
        add_climb(s)  # default climb calculator
        add_responses(s, c, self._ff_parameters(), prefix=N.DEFAULT)

    def _load_standard_statistics(self, s, c):
        add_statistics(s, ActivityCalculator, c, owner_in=short_cls(ResponseCalculator),
                       climb=CLIMB_CNAME, response_prefix=N.DEFAULT)
        add_statistics(s, SegmentCalculator, c, owner_in=short_cls(SegmentReader))
        add_statistics(s, MonitorCalculator, c, owner_in=short_cls(MonitorReader))
        add_statistics(s, RestHRCalculator, c, owner_in=short_cls(MonitorReader))
        add_statistics(s, KitCalculator, c, owner_in=short_cls(SegmentReader))
        add_statistics(s, SimilarityCalculator, c, owner_in=short_cls(ActivityCalculator))
        add_statistics(s, NearbyCalculator, c, owner_in=short_cls(SimilarityCalculator))

    def _load_summary_statistics(self, s, c):
        # need to call normalize here because schedule isn't a schedule type column,
        # but part of a kargs JSON blob.
        # also, add year first so that monthly doesn't get confused by extra stats range
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('x'))
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('y'))
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('m'))

    def _load_statistics_pipeline(self, s, c):
        # order is important here because some pipelines expect values created by others
        # this converts RAW_ELEVATION to ELEVATION, if needed
        add_statistics(s, ElevationCalculator, c)
        self._load_power_statistics(s, c)
        self._load_ff_statistics(s, c)
        self._load_standard_statistics(s, c)
        self._load_summary_statistics(s, c)
        add_statistics(s, AchievementCalculator, c, owner_in=short_cls(ActivityCalculator))

    def _load_diary_pipeline(self, s, c):
        add_displayer(s, DiaryDisplayer, c)
        add_displayer(s, MonitorDisplayer, c)
        # prefix ties in to the ff statistics config
        add_displayer(s, ResponseDisplayer, c, owner_in=short_cls(ResponseCalculator), prefix=N.DEFAULT)
        add_displayer(s, ActivityDisplayer, c)
        c2 = Counter()
        for delegate in self._activity_displayer_delegates():
            add_activity_displayer_delegate(s, delegate, c2)
        add_displayer(s, DatabaseDisplayer, c)

    def _activity_displayer_delegates(self):
        return [AchievementDelegate, ActivityDelegate, SegmentDelegate, NearbyDelegate, JupyterDelegate]

    def _load_monitor_pipeline(self, s, c):
        add_monitor(s, MonitorReader, c)

    def _load_constants(self, s):
        add_constant(s, SRTM1_DIR_CNAME,
                     base_system_path(self._base, version=PERMANENT, subdir='srtm1', create=False),
                     description='''
Directory containing SRTM1 hgt files for elevations.

These data are used to give improved values when using GPS elevation.
You must create the directory and populate it with suitable files from http://dwtkns.com/srtm30m.
If the directory or files are missing the raw GPS elevation will be used.
This is noted as a warning in the logs (along with the name of the missing file).
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)
        add_constant(s, GARMIN_USER, None,
                     description='''
User for Garmin.
If set, monitor data (daily steps, heart rate) are downloaded from Garmin.
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)
        add_constant(s, GARMIN_PASSWORD, None,
                     description='''
Password for Garmin.
This is stored as plaintext on the server (and visible here)
so do not use an important password that applies to many accounts.
''',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)

    def _load_diary_topics(self, s, c):
        # the fields you see every day in the diary
        diary = add_diary_topic(s, 'Status', c)
        add_diary_topic_field(s, diary, 'Notes', c, StatisticJournalType.TEXT,
                              model={TYPE: EDIT}, description='Daily notes recorded by user in diary.')
        add_diary_topic_field(s, diary, 'Weight', c, StatisticJournalType.FLOAT,
                              units='kg', summary=S.join(S.AVG, S.MSR),
                              description='Weight recorded by user in diary.',
                              model={TYPE: FLOAT, LO: 50, HI: 100, DP: 1})
        add_diary_topic_field(s, diary, 'Sleep', c, StatisticJournalType.FLOAT,
                              units='h', summary=S.AVG, description='Sleep time recorded by user in diary.',
                              model={TYPE: FLOAT, LO: 0, HI: 24, DP: 1})
        add_diary_topic_field(s, diary, 'Mood', c, StatisticJournalType.FLOAT,
                              summary=S.AVG, description='Mood recorded by user in diary.',
                              model={TYPE: SCORE})
        add_diary_topic_field(s, diary, 'Medication', c, StatisticJournalType.TEXT,
                              summary=S.CNT, description='Medication recorded by user in diary.',
                              model={TYPE: EDIT})
        add_diary_topic_field(s, diary, 'Weather', c, StatisticJournalType.TEXT,
                              summary=S.CNT, description='Weather recorded by user in diary.',
                              model={TYPE: EDIT})

    def _load_activity_topics(self, s, c):
        # the fields in the diary that are displayed for each activity
        for activity_group in self._activity_groups.values():
            c = Counter()
            root = add_activity_topic(s, ActivityTopic.ROOT, c, description=ActivityTopic.ROOT_DESCRIPTION,
                                      activity_group=activity_group)
            add_activity_topic_field(s, root, 'Name', c, StatisticJournalType.TEXT,
                                     activity_group, model={TYPE: EDIT},
                                     description=ActivityTopicField.NAME_DESCRIPTION)
            # note that these have empty toic parents because they are children of the entry itself
            if activity_group.name != SWIM:
                add_activity_topic_field(s, root, 'Route', c, StatisticJournalType.TEXT,
                                         activity_group, model={TYPE: EDIT},
                                         description='Route recorded by user in diary.')
            add_activity_topic_field(s, root, NOTES, c, StatisticJournalType.TEXT,
                                     activity_group, model={TYPE: EDIT},
                                     description='Activity notes recorded by user in diary.')

    def _load_sys(self, sys, s):
        # finally, update the timezone
        DiaryTopicJournal.check_tz(sys, s)




from logging import getLogger

from werkzeug import Response
from werkzeug.wrappers import ETagResponseMixin

from . import ContentType
from ...commands.args import base_system_path, THUMBNAIL
from ...commands.thumbnail import parse_activity, create_in_cache

log = getLogger(__name__)


class CacheResponse(Response, ETagResponseMixin):
    pass


class Thumbnail(ContentType):

    def __init__(self, base):
        self._base = base

    def __call__(self, request, s, activity):
        activity_id = parse_activity(s, activity)
        file = create_in_cache(self._base, s, activity_id)
        try:
            path = base_system_path(self._base, subdir=THUMBNAIL, file=file)
            log.debug(f'Reading {path}')
            with open(path, 'rb') as input:
                response = CacheResponse(input.read())
            self.set_content_type(response, file)
            response.cache_control.max_age = 3600
            return response
        except Exception as e:
            log.warning(f'Error serving {file}: {e}')
            raise



from json import dumps
from logging import getLogger

from ..names import simple_name
from ..sql import ActivityGroup, Constant, Pipeline, PipelineType, StatisticName, StatisticJournalType, \
    DiaryTopic, DiaryTopicField, Dummy, ActivityTopic, ActivityTopicField
from ..sql.tables.constant import ValidateNamedTuple
from ..sql.types import long_cls

log = getLogger(__name__)

NEARBY_CNAME = 'nearby'


class Counter:

    def __init__(self, start=10, delta=10):
        self.__start = start
        self.__delta = delta
        self.__previous = None

    def __call__(self, reset=None, delta=None):
        if delta is not None:
            if delta < 1:
                raise Exception('Negative increment')
            self.__delta = delta
        if reset is None:
            if self.__previous is None:
                self.__previous = self.__start
            else:
                self.__previous += self.__delta
        else:
            if reset <= self.__previous:
                raise Exception('Sort not increasing with reset')
            else:
                self.__previous = reset
        return self.__previous


def add(s, instance):
    '''
    Add an instance to the session (and so to the database), returning the instance.
    You likely don't need this - see the more specific helpers below.

    The instance can of any class that subclasses the Field class from SQLAlchemy.
    In practice, that means most classes in the ch2.sql.tables module.
    However, only some classes make sense in the context of a configuration, and
    more specific helpers probably already exist for those.
    '''
    s.add(instance)
    return instance


def add_activity_group(s, name, sort, description=None):
    '''
    Add an activity type to the configuration.

    These are used to group activities (and related statistics).
    So typical entries might be for cycling, running, etc.
    '''
    return add(s, ActivityGroup(name=name, sort=sort, description=description))


def add_statistics(s, cls, sort, **kargs):
    '''
    Add a class to the statistics pipeline.

    The pipeline classes are invoked when the diary is modified and when activities are added.
    They detect new data and calculate appropriate statistics.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    log.debug(f'Adding statistic pipeline {cls}')
    return add(s, Pipeline(cls=cls, type=PipelineType.CALCULATE, sort=sort, kargs=kargs))


def add_displayer(s, cls, sort, **kargs):
    '''
    Add a class to the diary pipeline.

    The pipeline classes are invoked when the diary is displayed.
    They generate display classes for activity statistics (and similar)

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    log.debug(f'Adding displayer pipeline {cls}')
    return add(s, Pipeline(cls=cls, type=PipelineType.DISPLAY, sort=sort, kargs=kargs))


def add_activity_displayer_delegate(s, cls, sort, **kargs):
    '''
    Add a class to the activity displayer pipeline.

    The pipeline classes are invoked when the diary is displayed, for each activity.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    log.debug(f'Adding activity displayer pipeline {cls}')
    return add(s, Pipeline(cls=cls, type=PipelineType.DISPLAY_ACTIVITY, sort=sort, kargs=kargs))


def add_monitor(s, cls, sort, **kargs):
    '''
    Add a class to the monitor pipeline.

    The pipeline classes are invoked when activities are imported from FIT files.
    They read the files and create MonitorJournal entries and associated statistics.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    log.debug(f'Adding monitor pipeline {cls}')
    return add(s, Pipeline(cls=cls, type=PipelineType.READ_MONITOR, sort=sort, kargs=kargs))


def add_activities(s, cls, sort, **kargs):
    '''
    Add a class to the activities pipeline.

    The pipeline classes are invoked when activities are imported from FIT files.
    They read the files and create ActivityJournal entries and associated statistics.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    log.debug(f'Loading activity pipeline {cls}')
    return add(s, Pipeline(cls=cls, type=PipelineType.READ_ACTIVITY, sort=sort, kargs=kargs))


def add_constant(s, title, value, description=None, units=None, name=None,
                 statistic_journal_type=StatisticJournalType.INTEGER, activity_group=None,
                 time=0.0, single=False, validate_cls=None, validate_args=None, validate_kargs=None):
    '''
    Add a constant.

    Configuring a constant allows the user to supply a value later, using the `ch2 constant` command.
    This can be useful for values that don't vary often, and so aren't worth adding to the diary.
    An example is FTHR, which you will only measure occasionally, but which is needed when calculating
    activity statistics (also, FTHR can vary by activity, which is why we add a constant per activity).
    '''
    if name is None: name = simple_name(title)
    log.debug(f'Adding constant {name}')
    statistic_name = StatisticName.add_if_missing(s, name, statistic_journal_type, units, None, Constant,
                                                  title=title, description=description)
    activity_group = ActivityGroup.from_name(s, activity_group)
    constant = add(s, Constant(statistic_name=statistic_name,
                               name=statistic_name.qualified_name(s, activity_group),
                               activity_group=activity_group, single=single, validate_cls=validate_cls,
                               validate_args=validate_args, validate_kargs=validate_kargs))
    if value:
        constant.add_value(s, value, time=time)
    else:
        log.warning(f'No value for constant {name}')
    return constant


def add_enum_constant(s, title, enum, value, description=None, units=None, single=False, name=None,
                      activity_group=None, time=0.0):
    '''
    Add a constant that is a JSON encoded enum.  This is validated before saving.
    '''
    return add_constant(s, title, dumps(value), description=description, units=units, name=name,
                        statistic_journal_type=StatisticJournalType.TEXT, activity_group=activity_group,
                        time=time, single=single, validate_cls=ValidateNamedTuple,
                        validate_args=[], validate_kargs={'tuple_cls': long_cls(enum)})


def set_constant(s, constant, value, time=None, date=None):
    '''
    Set a constant value.
    '''
    constant.add_value(s, value, time=time, date=date)


def add_diary_topic(s, title, sort, description=None, schedule=None):
    '''
    Add a root topic.

    DiaryTopics are displayed in the diary.
    They can be permanent, or associated with some schedule.
    They can also be associated with fields (and so with statistics).

    A root topic is usually used as a header to group related children.
    For example, 'DailyDiary' to group diary entries (notes, weight, sleep etc), or 'Plan' to group training plans.
    '''
    return add(s, DiaryTopic(title=title, sort=sort, description=description, schedule=schedule))


def add_child_diary_topic(s, parent, name, sort, description=None, schedule=None):
    '''
    Add a child topic.

    DiaryTopics are displayed in the diary.
    They can be permanent, or associated with some schedule.
    They can also be associated with fields (and so with statistics).

    A child topic is used to add additional structure to an existing topic.
    For example, the parent topic might be "injuries" and permanent, while children are defined for
    specific injuries with a schedule that gives start and end dates.
    '''
    return add(s, DiaryTopic(parent=parent, title=name, sort=sort, description=description, schedule=schedule))


def add_diary_topic_field(s, diary_topic, name, sort, type, description=None, units=None, summary=None, schedule=None,
                          model=None):
    '''
    Add a field and associated statistic to a topic entry.

    This is how the user can enter values into the diary.
    The field describes how the values are displayed in the diary.
    The statistic describes how the values are stored in the database.
    '''
    if diary_topic.id is None:
        s.flush()
    statistic_name = add(s, StatisticName(name=name, owner=DiaryTopic, statistic_journal_type=type,
                                          description=description, units=units, summary=summary))
    if model is None: model = {}
    field = add(s, DiaryTopicField(diary_topic=diary_topic, sort=sort, model=model, schedule=schedule,
                                   statistic_name=statistic_name))


def add_activity_topic(s, title, sort, description=None, activity_group=None):
    '''
    Add a root topic.

    DiaryTopics are displayed in the diary.
    They can be permanent, or associated with some schedule.
    They can also be associated with fields (and so with statistics).

    A root topic is usually used as a header to group related children.
    For example, 'DailyDiary' to group diary entries (notes, weight, sleep etc), or 'Plan' to group training plans.
    '''
    return add(s, ActivityTopic(title=title, sort=sort, description=description,
                                activity_group=ActivityGroup.from_name(s, activity_group)))


def add_child_activity_topic(s, parent, title, sort, description=None):
    '''
    Add a child topic.

    DiaryTopics are displayed in the diary.
    They can be permanent, or associated with some schedule.
    They can also be associated with fields (and so with statistics).

    A child topic is used to add additional structure to an existing topic.
    For example, the parent topic might be "injuries" and permanent, while children are defined for
    specific injuries with a schedule that gives start and end dates.
    '''
    return add(s, ActivityTopic(parent=parent, title=title, sort=sort, description=description))


def add_activity_topic_field(s, activity_topic, title, sort, type, activity_group,
                             description=None, units=None, summary=None, model=None):
    '''
    Add a field and associated statistic to a topic entry.

    This is how the user can enter values into the diary.
    The field describes how the values are displayed in the diary.
    The statistic describes how the values are stored in the database.
    '''
    if activity_topic and activity_topic.id is None:
        s.flush()
    # cannot simply add as this is also called during loading
    statistic_name = StatisticName.add_if_missing(s, title, type, units, summary, ActivityTopic,
                                                  description=description)
    if model is None: model = {}
    return add(s, ActivityTopicField(activity_topic=activity_topic, sort=sort, model=model,
                                     statistic_name=statistic_name))


def add_loader_support(s):
    '''
    Add 'dummy' value used by loader.
    '''
    log.debug('Adding dummy source')
    dummy_source = add(s, Dummy())
    dummy_name = add(s, StatisticName(name=Dummy.DUMMY, owner=dummy_source,
                                      statistic_journal_type=StatisticJournalType.STATISTIC))




from sqlalchemy import Column, Integer, ForeignKey, Float, Text, UniqueConstraint, or_
from sqlalchemy.orm import relationship

from .source import SourceType, Source, GroupedSource
from ..support import Base
from ..types import Time
from ...lib.date import format_time


class SegmentJournal(GroupedSource):

    __tablename__ = 'segment_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    # do not use on delete cascade here since it leaves source entries without children
    # instead, to guarantee consistency, call clean()
    segment_id = Column(Integer, ForeignKey('segment.id', ondelete='set null'), index=True)
    segment = relationship('Segment')
    # do not use on delete cascade here since it leaves source entries without children
    # instead, to guarantee consistency, call clean()
    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='set null'),
                                 index=True)
    activity_journal = relationship('ActivityJournal', foreign_keys=[activity_journal_id])
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(segment_id, activity_journal_id, start, finish)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SEGMENT
    }

    def __str__(self):
        return 'Segment Journal %s to %s' % (format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish

    @classmethod
    def clean(cls, s):
        q1 = s.query(SegmentJournal.id). \
                filter(or_(SegmentJournal.segment_id == None,
                           SegmentJournal.activity_journal_id == None)).cte()
        s.query(Source).filter(Source.id.in_(q1)).delete(synchronize_session=False)


class Segment(Base):

    __tablename__ = 'segment'

    id = Column(Integer, primary_key=True)
    # segments used to depend on activity group.  but then they also depended on equipment used,
    # so now they're agnostic about other tables and you need to join and filter as appropriate when extracting
    start_lat = Column(Float, nullable=False)
    start_lon = Column(Float, nullable=False)
    finish_lat = Column(Float, nullable=False)
    finish_lon = Column(Float, nullable=False)
    distance = Column(Float, nullable=False)
    title = Column(Text, nullable=False, index=True, unique=True)
    description = Column(Text)
    UniqueConstraint(start_lat, start_lon, finish_lat, finish_lon)

    @property
    def start(self):
        return self.start_lon, self.start_lat

    @start.setter
    def start(self, lon_lat):
        self.start_lon, self.start_lat = lon_lat

    @property
    def finish(self):
        return self.finish_lon, self.finish_lat

    @finish.setter
    def finish(self, lon_lat):
        self.finish_lon, self.finish_lat = lon_lat

    def coords(self, start):
        if start:
            return self.start
        else:
            return self.finish

    def __str__(self):
        return f'Segment "{self.title}"'



from logging import getLogger
from re import compile

import numpy as np
import pandas as pd
from sqlalchemy import inspect

from .coasting import CoastingBookmark
from ..lib.data import kargs_to_attr
from ..names import Names as N, like
from ..sql import StatisticName, StatisticJournal, StatisticJournalInteger, ActivityJournal, \
    StatisticJournalFloat, StatisticJournalText, Interval, Source
from ..sql.database import connect, ActivityTimespan, ActivityBookmark, Composite, \
    CompositeComponent, ActivityNearby

log = getLogger(__name__)



def read_query(query, index=None):
    '''
    Convert s.query(OrmClass) to a dataframe
    https://stackoverflow.com/questions/29525808/sqlalchemy-orm-conversion-to-pandas-dataframe
    '''
    return pd.read_sql(query.statement, query.session.bind, index_col=index)


def session(*args):
    '''
    Create a database session (used in Jupyter templates)
    '''
    ns, db = connect(args)
    return db.session()


def _tables():
    return kargs_to_attr(sj=inspect(StatisticJournal).local_table,
                         sn=inspect(StatisticName).local_table,
                         sji=inspect(StatisticJournalInteger).local_table,
                         sjf=inspect(StatisticJournalFloat).local_table,
                         sjt=inspect(StatisticJournalText).local_table,
                         inv=inspect(Interval).local_table,
                         at=inspect(ActivityTimespan).local_table,
                         aj=inspect(ActivityJournal).local_table,
                         cmp=inspect(Composite).local_table,
                         cc=inspect(CompositeComponent).local_table,
                         src=inspect(Source).local_table)


def nearby_activities(s, local_time=None, activity_group=None):
    from ..pipeline.display.activity.nearby import nearby_any_time
    journal = ActivityJournal.at_local_time(s, local_time, activity_group=activity_group)
    return nearby_any_time(s, journal)


def bookmarks(s, constraint, owner=CoastingBookmark):
    yield from s.query(ActivityBookmark). \
        filter(ActivityBookmark.owner == owner,
               ActivityBookmark.constraint == constraint).all()


def present(df, *names, pattern=False):
    if pattern:
        if hasattr(df, 'columns'):
            for name in names:
                columns = like(name, df.columns)
                if not columns or not all(len(df[column].dropna()) for column in columns):
                    return False
            return True
        else:
            return df is not None and (len(df.dropna()) and all(like(name, [df.name]) for name in names))
    else:
        if hasattr(df, 'columns'):
            return all(name in df.columns and len(df[name].dropna()) for name in names)
        else:
            return df is not None and (len(df.dropna()) and all(df.name == name for name in names))


def median_d(df):
    return pd.Series(df.index).diff().median()


KEEP = 'keep'


def linear_resample(df, start=None, finish=None, d=None, quantise=True):
    log.debug(f'Linear resample with index {type(df.index)}, columns {df.columns}')
    d = d or median_d(df)
    start = start or df.index.min()
    finish = finish or df.index.max()
    if quantise:
        start, finish = int(start / d) * d, (1 + int(finish / d)) * d
    lin = pd.DataFrame({KEEP: True}, index=np.arange(start, finish, d))
    ldf = df.join(lin, how='outer', sort=True)
    # if this fails check for time-like columns
    ldf.interpolate(method='slinear', limit_area='inside', inplace=True)
    ldf = ldf.loc[ldf[KEEP] == True].drop(columns=[KEEP])
    return ldf


def median_dt(df):
    return pd.Series(df.index).diff().median().total_seconds()


def linear_resample_time(df, start=None, finish=None, dt=None, with_timespan=False, keep_nan=True, add_time=True):
    log.debug(f'Linear resample with index {type(df.index)}, columns {df.columns}')
    if with_timespan is None: with_timespan = N.TIMESPAN_ID in df.columns
    dt = dt or median_dt(df)
    start = start or df.index.min()
    finish = finish or df.index.max()
    lin = pd.DataFrame({KEEP: True}, index=pd.date_range(start=start, end=finish, freq=f'{dt}S'))
    ldf = df.join(lin, how='outer', sort=True)
    # if this fails check for time-like columns
    ldf.interpolate(method='index', limit_area='inside', inplace=True)
    ldf = ldf.loc[ldf[KEEP] == True].drop(columns=[KEEP])
    if add_time:
        ldf[N.TIME] = ldf.index
        ldf[N.DELTA_TIME] = ldf[N.TIME].diff()
    if with_timespan:
        if keep_nan:
            ldf.loc[~ldf[N.TIMESPAN_ID].isin(df[N.TIMESPAN_ID].unique())] = np.nan
        else:
            ldf = ldf.loc[ldf[N.TIMESPAN_ID].isin(df[N.TIMESPAN_ID].unique())]
    return ldf


def groups_by_time(s, start=None, finish=None):
    q = s.query(ActivityJournal.start.label(N.INDEX), ActivityNearby.group.label(N.GROUP)). \
        filter(ActivityNearby.activity_journal_id == ActivityJournal.id)
    if start:
        q = q.filter(ActivityJournal.start >= start)
    if finish:
        q = q.filter(ActivityJournal.start < finish)
    return read_query(q, index=N.INDEX)


def related_statistics(df, statistic, unpack=r'({statistic})\s*:\s*([^:]+)'):
    rx = compile(unpack.format(statistic=statistic))
    for column in df.columns:
        m = rx.match(column)
        if m: yield m.group(1), m.group(2)


def transform(df, transformation):
    transformation = {key: value for key, value in transformation.items() if key in df.columns}
    return df.transform(transformation)


def drop_empty(df):
    for column in df.columns:
        if df[column].dropna().empty:
            df = df.drop(columns=[column])
    return df


def valid(value):
    return value == value and value is not None



from logging import getLogger

import sqlalchemy as s
from sqlalchemy.orm import sessionmaker

from ch2.names import SPACE, simple_name
from ch2.sql.support import Base
from ch2.sql.types import Date
from tests import LogTestCase

log = getLogger(__name__)


class Data(Base):

    __tablename__ = 'data'

    integer = s.Column(s.Integer, primary_key=True)
    text = s.Column(s.Text, nullable=False, server_default='')
    date = s.Column(Date)


class Database:

    def __init__(self):
        self.engine = s.create_engine('sqlite:///:memory:', echo=True)
        self.__create_tables()
        self.session = sessionmaker(bind=self.engine)

    def __create_tables(self):
        log.info('Creating tables')
        Base.metadata.create_all(self.engine)


class TestName(LogTestCase):

    def test_tokenzie(self):
        self.assertEqual(simple_name('ABC 123 *^%'), 'abc-123-%')  # support for like
        self.assertEqual(simple_name('****'), SPACE)
        self.assertEqual(simple_name('123'), '-123')
        self.assertEqual(simple_name('Fitness 7d'), 'fitness-7d')




import datetime as dt
import re
from logging import getLogger

import numpy as np
from math import exp
from pandas import DataFrame, Series
from scipy import optimize

from .lib import decay_params, inplace_decay
from ..lib.data import safe_dict
from ..names import Titles, like, N, SPACE

log = getLogger(__name__)

IMPULSE_3600 = 'Impulse / 3600s'
RESPONSE = 'Response'
LOG10_PERIOD, LOG10_START = 0, 1


# Almost everything below uses Series, not DataFrame
# Currently model params are log10_period (hours) and log10_start (initial FF value)


def sum_to_hour(source, column):
    data = source.resample('1h', label='right').sum()
    data = data.loc[:, [column]]
    data.rename(columns={column: IMPULSE_3600}, inplace=True)
    initial = DataFrame({IMPULSE_3600: 0}, index=[data.index[0] - dt.timedelta(hours=1)])
    data = initial.append(data, sort=True)
    return data


def calc_response(data, params):
    # data is a DataFrame mainly because that's what inplace_decay takes
    response = data.rename(columns={IMPULSE_3600: RESPONSE})  # copy
    period = 10 ** params[LOG10_PERIOD]
    if len(params) > LOG10_START:
        decay, alpha = decay_params(period)
        # correct by alpha so that the value matches the plot (see decay_inplace)
        start = 10 ** params[LOG10_START] * alpha
        # we replace the very first value which is strictly cheating, but there are so many...
        response.at[response.index[0], RESPONSE] = start
    inplace_decay(response, RESPONSE, period)
    return response[RESPONSE]


def restrict_response(response, performances):
    # we're only interested in the FF model at the times that correspond to performance measurements
    return [response.reindex(index=performance.index, method='nearest')
            for performance in performances]


def calc_observed(measureds, performances):
    # this is a bit tricky.  we don't know the exact relationship between 'fitness' and how fast we are.
    # all we know is that it should track changes in a useful manner.
    # so there's some arbitrary transform, which could quite easily be different for different routes
    # (no reason 'fitness' should numerically predict the speed on *all* routes, obviously).
    # so we assume that there's an arbitrary linear transform (scaling and offset) from measured speed
    # (or whatever 'performance' is) to 'fitness'.
    # we could fit for the best transform as we fit for the period, but it's more efficient to calculate
    # the 'best' transform for a given period.  this is more efficient because it's equivalent to line fitting
    # (that linear transform is y = ax + b).  so we can call the (fast) line fitting routine in numpy.
    return [Series(np.polyval(np.polyfit(performance, measured, 1), performance),
                   performance.index)
            for measured, performance in zip(measureds, performances)]


def calc_residuals(models, observations, method='L1'):
    # -ve when observation less than model
    if method == 'L1':
        # use L1 scaled by amplitude to be reasonably robust.
        return [(observation - model) / model.clip(lower=1e-6)
                for model, observation in zip(models, observations)]
    elif method == 'L2':
        # something like chisq
        return [(np.sign(observation - model) * (observation - model).pow(2) / model.clip(lower=1e-6))
                for model, observation in zip(models, observations)]
    else:
        raise Exception(f'Unknown method ({method}) - use L1 or L2')


def calc_cost(model, observation, method='L1'):
    residuals = calc_residuals(model, observation, method=method)
    return sum(sum(np.abs(residual)) for residual in residuals)


def worst_residual(residuals, threshold=None):
    i, t, worst = None, None, None
    for index, residual in enumerate(residuals):
        bad = max(residual)
        if bad > threshold and (worst is None or bad > worst):
            worst, i, t = bad, index, residual.idxmax()
    if worst:
        log.debug(f'Worst residual magnitude {worst} at {t}')
    else:
        log.debug('No worst residual found')
    return i, t, worst


def fix_residuals(residuals, threshold=None):
    try:
        pos, neg = np.abs(threshold)
        log.debug(f'Scaling residuals to weight lower by {pos}:{neg}')
        for residual in residuals:
            residual.loc[residual < 0] = residual.loc[residual < 0].abs() * pos / neg
        return residuals, pos
    except TypeError:
        return [residual.abs() for residual in residuals], 0 if threshold is None else threshold


def reject_worst_inplace(params, data, performances, method='L1', threshold=None):
    response = calc_response(data, params)
    measureds = restrict_response(response, performances)
    predicteds = calc_observed(measureds, performances)
    residuals = calc_residuals(measureds, predicteds, method=method)
    residuals, threshold = fix_residuals(residuals, threshold=threshold)
    i, t, worst = worst_residual(residuals, threshold=threshold)
    if t:
        print(f'Dropping {worst} at {t}')
        performances[i].drop(index=t, inplace=True)
    return i, t


def fmt_params(params):
    text = f'Period {10 ** params[0] / 24:.1f}'
    if len(params) > 1:
        text += f'; Start {10 ** params[1]:.1f}'
    return text


def fit_ff_params(data, params, performances, method='L1', max_reject=0, threshold=None, **kargs):
    # data should be a DataFrame with an IMPULSE3600 entry
    # performances should be Series
    # threshold can be None, a value, or a pair.  a pair gives +ve and -ve thresholds

    result, rejected = None, []

    def cost(params):
        response = calc_response(data, params)
        restricted = restrict_response(response, performances)
        observed = calc_observed(restricted, performances)
        return calc_cost(restricted, observed, method=method)

    while True:
        result = optimize.minimize(cost, params, **kargs)
        print(fmt_params(result.x))
        print(f'Currently have {len(rejected)} dropped points (max {max_reject})')
        if len(rejected) >= max_reject:
            return result, rejected
        i, t = reject_worst_inplace(params, data, performances, method=method, threshold=threshold)
        if t is None:
            return result, rejected
        rejected.append((i, t))


@safe_dict
def response_stats(df, prev_secs):
    from math import log
    digits = re.compile(r'(\d+)')
    stats = {}
    for pattern in (N.FITNESS_ANY, N.FATIGUE_ANY):
        for name in like(pattern, df.columns):
            days = int(digits.search(name).group(1))
            lower, higher = df[name][0], df[name][-1]
            delta = higher - lower
            stats[N._delta(name)] = delta
            tau = days * 24 * 60 * 60
            # this is the time needed for the value to return to where it was before the activity
            # so for fitness it's kinda the time 'bought' within which you're not getting worse and
            # for fatigue it's the recovery time.
            revert = tau * log(1 + delta / lower)
            if N.FITNESS in name:
                stats[N.EARNED_D % days] = revert
                if prev_secs:
                    # this was an experiment.  if you exercise regularly at the same intensity then you
                    # will tend to a certain fitness level.  this is an estimate of that level, assuming
                    # that activities repeat at the same interval as the time from the previous activity.
                    # it was a cute idea, but turns out to be way too noisy to be useful.
                    plateau = delta / (exp(prev_secs / tau) - 1)
                    stats[N.PLATEAU_D % days] = plateau
            else:
                stats[N.RECOVERY_D % days] = revert
    return stats



from logging import getLogger

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

from .utils import MultiProcCalculator, IntervalCalculatorMixin
from ..pipeline import OwnerInMixin
from ...data import Statistics
from ...lib import format_date, local_date_to_time
from ...names import Titles, Summaries as S, Units, Names
from ...sql import StatisticJournalInteger

log = getLogger(__name__)


class RestHRCalculator(OwnerInMixin, IntervalCalculatorMixin, MultiProcCalculator):

    '''
    Used to calculate rest HR from quartiles, but it was never clear we had *the* rest value rather
    than some general lower value.

    This way - by looking for the first peak that's not noise - we are finding something that perhaps
    is more meaningful.  It's a low heart rate that you spent a fair amount of time at.
    '''

    def __init__(self, *args, schedule='d', **kargs):
        super().__init__(*args, schedule=schedule, **kargs)

    def _read_data(self, s, interval):
        return Statistics(s, start=local_date_to_time(interval.start),
                          finish=local_date_to_time(interval.finish)). \
            by_name(self.owner_in, Names.HEART_RATE).df

    def _calculate_results(self, s, interval, df, loader):
        if not df.empty:
            hist = pd.cut(df[Names.HEART_RATE], np.arange(30, 90), right=False).value_counts(sort=False)
            peaks, _ = find_peaks(hist)
            for peak in peaks:
                rest_hr = hist.index[peak].left
                measurements = hist.loc[rest_hr]
                if measurements > len(df) * 0.01:
                    log.debug(f'Rest HR is {rest_hr} with {measurements} values')
                    # conversion to int as value above is numpy int64
                    loader.add(Titles.REST_HR, Units.BPM, S.join(S.MIN, S.MSR), interval,
                               int(rest_hr), local_date_to_time(interval.start), StatisticJournalInteger,
                               'The rest heart rate')
                    return
                else:
                    log.debug(f'Skipping rest HR at {rest_hr} because too few measurements ({measurements}/{len(df)})')
        log.warning(f'Unable to calculate rest HR at {format_date(interval.start)}')





import datetime as dt
from collections import defaultdict
from json import loads
from logging import getLogger

from sqlalchemy import desc

from .elevation import ElevationCalculator
from .impulse import ImpulseCalculator
from .power import PowerCalculator
from .utils import MultiProcCalculator, ActivityJournalCalculatorMixin, DataFrameCalculatorMixin
from ..pipeline import OwnerInMixin
from ..read.segment import SegmentReader
from ...data import Statistics
from ...data.activity import active_stats, times_for_distance, hrz_stats, max_med_stats, max_mean_stats, \
    direction_stats, copy_times
from ...data.climb import find_climbs, Climb, add_climb_stats
from ...data.frame import present
from ...data.response import response_stats
from ...lib import time_to_local_time
from ...lib.data import safe_dict
from ...lib.log import log_current_exception
from ...names import N, T, Summaries as S, Units, titles_for_names, SPACE, simple_name
from ...sql import StatisticJournalFloat, Constant, StatisticJournalText, StatisticJournalTimestamp, \
    ActivityJournal, StatisticJournal

log = getLogger(__name__)


class ActivityCalculator(OwnerInMixin, ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, climb=None, response_prefix=None, **kargs):
        self.climb_ref = climb
        self.response_prefix = response_prefix
        super().__init__(*args, add_serial=False, **kargs)

    def _read_dataframe(self, s, ajournal):
        try:
            adf = Statistics(s, activity_journal=ajournal, with_timespan=True). \
                by_name(SegmentReader, N.DISTANCE, N.HEART_RATE, N.SPHERICAL_MERCATOR_X, N.SPHERICAL_MERCATOR_Y). \
                by_name(ElevationCalculator, N.ELEVATION). \
                by_name(ImpulseCalculator, N.HR_ZONE). \
                by_name(PowerCalculator, N.POWER_ESTIMATE).df
            if self.response_prefix:
                start, finish = ajournal.start - dt.timedelta(hours=1), ajournal.finish + dt.timedelta(hours=1)
                fitness = self.response_prefix + SPACE + N.FITNESS_ANY
                fatigue = self.response_prefix + SPACE + N.FATIGUE_ANY
                sdf = Statistics(s, start=start, finish=finish). \
                    by_name(self.owner_in, fitness, fatigue, like=True).with_. \
                    drop_prefix(self.response_prefix).df
            else:
                sdf = None
            prev = s.query(ActivityJournal). \
                filter(ActivityJournal.start < ajournal.start). \
                order_by(desc(ActivityJournal.start)). \
                first()
            delta = (ajournal.start - prev.start).total_seconds() if prev else None
            return adf, sdf, delta
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise

    def _calculate_stats(self, s, ajournal, data):
        adf, sdf, delta = data
        stats, climbs = {}, None
        stats.update(copy_times(ajournal))
        stats.update(active_stats(adf))
        stats.update(self.__average_power(s, ajournal, stats[N.ACTIVE_TIME]))
        stats.update(times_for_distance(adf))
        stats.update(hrz_stats(adf))
        stats.update(max_med_stats(adf))
        stats.update(max_mean_stats(adf))
        stats.update(direction_stats(adf))
        if sdf is not None:
            stats.update(response_stats(sdf, delta))
        if present(adf, N.ELEVATION):
            params = Climb(**loads(Constant.from_name(s, self.climb_ref).at(s).value))
            climbs = list(find_climbs(adf, params=params))
            add_climb_stats(adf, climbs)
        return data, stats, climbs

    @safe_dict
    def __average_power(self, s, ajournal, active_time):
        # this doesn't fit nicely anywhere...
        energy = StatisticJournal.at(s, ajournal.start, N.ENERGY_ESTIMATE, PowerCalculator,
                                     ajournal.activity_group)
        if energy and active_time:
            return {N.MEAN_POWER_ESTIMATE: 1000 * energy.value / active_time}
        else:
            return {N.MEAN_POWER_ESTIMATE: 0}

    def _copy_results(self, s, ajournal, loader, data):
        df, stats, climbs = data
        self.__copy(ajournal, loader, stats, T.START, None,
                    None, ajournal.start, type=StatisticJournalTimestamp)
        self.__copy(ajournal, loader, stats, T.FINISH, None,
                    None, ajournal.start, type=StatisticJournalTimestamp)
        self.__copy(ajournal, loader, stats, T.TIME, Units.S,
                    S.join(S.MAX, S.SUM, S.MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, T.ACTIVE_DISTANCE, Units.KM,
                    S.join(S.MAX, S.CNT, S.SUM, S.MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, T.ACTIVE_TIME, Units.S,
                    S.join(S.MAX, S.SUM, S.MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, T.ACTIVE_SPEED, Units.KMH,
                    S.join(S.MAX, S.AVG, S.MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, T.MEAN_POWER_ESTIMATE,Units. W,
                    S.join(S.MAX, S.AVG, S.MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, T.DIRECTION, Units.DEG,
                    None, ajournal.start)
        self.__copy(ajournal, loader, stats, T.ASPECT_RATIO, None,
                    None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.MIN_KM_TIME_ANY, Units.S,
                        S.join(S.MIN, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.MED_KM_TIME_ANY, Units.S,
                        S.join(S.MIN, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.PERCENT_IN_Z_ANY, Units.PC,
                        None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.TIME_IN_Z_ANY, Units.S,
                        None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.MAX_MED_HR_M_ANY, Units.BPM,
                        S.join(S.MAX, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.MAX_MEAN_PE_M_ANY, Units.W,
                        S.join(S.MAX, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T._delta(T.FITNESS_ANY), Units.FF,
                        S.join(S.MAX, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T._delta(T.FATIGUE_ANY), Units.FF,
                        S.join(S.MAX, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.EARNED_D_ANY, Units.S,
                        S.join(S.MAX, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.RECOVERY_D_ANY, Units.S,
                        S.join(S.MAX, S.MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, T.PLATEAU_D_ANY, Units.FF,
                        None, ajournal.start)
        if climbs:
            loader.add(T.TOTAL_CLIMB, Units.M, S.join(S.MAX, S.MSR), ajournal,
                       sum(climb[N.CLIMB_ELEVATION] for climb in climbs), ajournal.start, StatisticJournalFloat,
                       description=DESCRIPTIONS[T.TOTAL_CLIMB])
            for climb in sorted(climbs, key=lambda climb: climb[N.TIME]):
                self.__copy(ajournal, loader, climb, T.CLIMB_ELEVATION, Units.M,
                            S.join(S.MAX, S.SUM, S.MSR), climb[N.TIME])
                self.__copy(ajournal, loader, climb, T.CLIMB_DISTANCE, Units.KM,
                            S.join(S.MAX, S.SUM, S.MSR), climb[N.TIME])
                self.__copy(ajournal, loader, climb, T.CLIMB_TIME, Units.S,
                            S.join(S.MAX, S.SUM, S.MSR), climb[N.TIME])
                self.__copy(ajournal, loader, climb, T.CLIMB_GRADIENT, Units.PC,
                            S.join(S.MAX, S.MSR), climb[N.TIME])
                self.__copy(ajournal, loader, climb, T.CLIMB_POWER,Units. W,
                            S.join(S.MAX, S.MSR), climb[N.TIME])
                if T.CLIMB_CATEGORY in climb:
                    self.__copy(ajournal, loader, climb, T.CLIMB_CATEGORY, None,
                                None, climb[T.TIME], type=StatisticJournalText)
        if stats:
            log.warning(f'Unsaved statistics: {list(stats.keys())}')

    def __copy_all(self, ajournal, loader, stats, pattern, units, summary, time, type=StatisticJournalFloat):
        description = DESCRIPTIONS[pattern]
        for title in list(titles_for_names(pattern, stats)):  # list to avoid reading after modification
            self.__copy(ajournal, loader, stats, title, units, summary, time, type=type, description=description)

    def __copy(self, ajournal, loader, stats, title, units, summary, time, type=StatisticJournalFloat,
               description=None):
        if not description: description = DESCRIPTIONS[title]
        name = simple_name(title)
        if name in stats:
            try:
                loader.add(title, units, summary, ajournal, stats[name], time, type, description=description)
            except:
                log.warning(f'Failed to load {title}')
                log_current_exception(traceback=False)
            del stats[name]
        else:
            log.warning(f'Did not calculate {title} '
                        f'({time_to_local_time(ajournal.start)} / {ajournal.activity_group.name})')


DESCRIPTIONS = defaultdict(lambda: None, {
    T.START: '''The start time for the activity.''',
    T.FINISH: '''The finish time for the activity.''',
    T.TIME: '''The total duration of the activity.''',
    T.ACTIVE_DISTANCE: '''The total distance travelled while active (ie not paused).''',
    T.ACTIVE_TIME: '''The total time while active (ie not paused).''',
    T.ACTIVE_SPEED: '''The average speed while active (ie not paused).''',
    T.MEAN_POWER_ESTIMATE: '''The average estimated power.''',
    T.DIRECTION: '''The angular direction (clockwise from North) of the mid-point of the activity relative to the start.''',
    T.ASPECT_RATIO: f'''The relative extent of the activity along and across the {T.DIRECTION}.''',
    T.MIN_KM_TIME_ANY: '''The shortest time required to cover the given distance.''',
    T.MED_KM_TIME_ANY: '''The median (typical) time required to cover the given distance.''',
    T.PERCENT_IN_Z_ANY: '''The percentage of time in the given HR zone.''',
    T.TIME_IN_Z_ANY: '''The total time in the given HR zone.''',
    T.MAX_MED_HR_M_ANY: '''The highest median HR in the given interval.''',
    T.MAX_MEAN_PE_M_ANY: '''The highest average power estimate in the given interval.''',
    T._delta(T.FATIGUE_ANY): '''The change (over the activity) in the SHRIMP Fatigue parameter.''',
    T._delta(T.FITNESS_ANY): '''The change (over the activity) in the SHRIMP Fitness parameter.''',
    T.EARNED_D_ANY: '''The time before Fitness returns to the value before the activity.''',
    T.RECOVERY_D_ANY: '''The time before Fatigue returns to the value before the activity.''',
    T.PLATEAU_D_ANY: '''The maximum Fitness achieved if this activity was repeated (with the same time gap to the previous).''',
    T.TOTAL_CLIMB: '''The total height climbed in the detected climbs (only).''',
    T.CLIMB_ELEVATION: '''The difference in elevation between start and end of the climb.''',
    T.CLIMB_DISTANCE: '''The distance travelled during the climb''',
    T.CLIMB_TIME: '''The time spent on the climb.''',
    T.CLIMB_GRADIENT: '''The average inclination of the climb (elevation / distance).''',
    T.CLIMB_POWER: '''The average estimated power during the climb.''',
    T.CLIMB_CATEGORY: '''The climb category (text, "4" to "1" and "HC").'''
})